from security.backends.app import SecurityBackend


class SecurityElasticsearchBackend(SecurityBackend):

    name = 'security.backends.elasticsearch'
    label = 'security_backends_elasticsearch'
    backend_name = 'elasticsearch'
    writer = 'security.backends.elasticsearch.writer.ElasticsearchBackendWriter'
    reader = 'security.backends.elasticsearch.reader.ElasticsearchBackendReader'



from django.core.exceptions import ObjectDoesNotExist
from django.core.serializers.json import DjangoJSONEncoder
from django.contrib.auth import get_user_model
from django.contrib.contenttypes.models import ContentType
from django.db import models
from django.utils.encoding import force_text
from django.utils.functional import cached_property
from django.utils.timezone import localtime

from generic_m2m_field.models import MultipleDBGenericManyToManyField

from chamber.models import SmartQuerySet

from enumfields import IntegerEnumField

from security.backends.common.mixins import (
    CeleryTaskInvocationLogStrMixin, CeleryTaskRunLogStrMixin, CommandLogStrMixin, InputRequestLogStrMixin,
    LogShortIdMixin, OutputRequestLogStrMixin
)
from security.enums import (
    LoggerName, RequestLogState, CeleryTaskInvocationLogState, CeleryTaskRunLogState, CommandState
)


class BaseLogQuerySet(SmartQuerySet):

    def filter_related_with_object(self, related_object):
        return self.filter(
            related_objects__object_id=related_object.pk,
            related_objects__object_ct_id=ContentType.objects.get_for_model(related_object).pk
        )


class Log(LogShortIdMixin, models.Model):

    id = models.UUIDField(
        null=False,
        blank=False,
        max_length=250,
        primary_key=True
    )
    extra_data = models.JSONField(
        null=True,
        blank=True,
        encoder=DjangoJSONEncoder
    )
    slug = models.CharField(
        null=True,
        blank=True,
        db_index=True,
        max_length=255
    )
    start = models.DateTimeField(
        blank=False,
        null=False,
        editable=False
    )
    stop = models.DateTimeField(
        blank=True,
        null=True,
        editable=False
    )
    time = models.FloatField(
        null=True,
        blank=True
    )
    release = models.CharField(
        null=True,
        blank=True,
        db_index=True,
        max_length=255
    )
    parent_log = models.CharField(
        null=True,
        blank=True,
        db_index=True,
        max_length=250
    )
    error_message = models.TextField(
        null=True,
        blank=True
    )
    version = models.PositiveSmallIntegerField(
        null=False,
        blank=False
    )

    def save(self, update_fields=None, *args, **kwargs):
        if not self.time and self.start and self.stop:
            self.time = (self.stop - self.start).total_seconds()
            if update_fields:
                update_fields = list(update_fields) + ['time']
        super().save(update_fields=update_fields, *args, **kwargs)

    class Meta:
        abstract = True


class RequestLog(Log):

    host = models.CharField(
        max_length=255,
        null=False,
        blank=False,
        db_index=True
    )
    method = models.SlugField(
        max_length=7,
        null=False,
        blank=False,
        db_index=True
    )
    path = models.CharField(
        max_length=2000,
        null=False,
        blank=True,
        db_index=True
    )
    queries = models.JSONField(
        null=True,
        blank=True,
        encoder=DjangoJSONEncoder
    )
    is_secure = models.BooleanField(
        default=False,
        null=False,
        blank=False
    )

    # Request information
    request_headers = models.JSONField(
        null=True,
        blank=True,
        encoder=DjangoJSONEncoder
    )
    request_body = models.TextField(
        null=True,
        blank=True
    )

    # Response information
    response_code = models.PositiveSmallIntegerField(
        null=True,
        blank=True
    )
    response_headers = models.JSONField(
        null=True,
        blank=True,
        encoder=DjangoJSONEncoder
    )
    response_body = models.TextField(
        null=True,
        blank=True
    )

    state = IntegerEnumField(
        enum=RequestLogState,
        null=False,
        blank=False,
        default=RequestLogState.INCOMPLETE
    )

    def __str__(self):
        return ' '.join(
            (force_text(v) for v in (
                self.slug, self.response_code, localtime(self.start.replace(microsecond=0))
            ) if v)
        )

    class Meta:
        abstract = True


class InputRequestLog(InputRequestLogStrMixin, RequestLog):

    user_id = models.TextField(
        null=True,
        blank=True,
        db_index=True
    )
    ip = models.GenericIPAddressField(
        null=False,
        blank=False,
        db_index=True
    )
    view_slug = models.CharField(
        null=True,
        blank=True,
        db_index=True,
        max_length=255
    )
    related_objects = MultipleDBGenericManyToManyField()

    objects = BaseLogQuerySet.as_manager()

    class Meta:
        ordering = ('-start',)

    @cached_property
    def user(self):
        return get_user_model().objects.filter(pk=self.user_id).first()


class OutputRequestLog(OutputRequestLogStrMixin, RequestLog):

    related_objects = MultipleDBGenericManyToManyField()

    objects = BaseLogQuerySet.as_manager()

    class Meta:
        ordering = ('-start',)

    class SmartMeta:
        is_cleaned_pre_save = False


class CommandLog(CommandLogStrMixin, Log):
    """
    Represents a log of a command run.

    Attributes:
        start: Date and time when command was started.
        stop: Date and time when command finished.
        time: Command processing time in miliseconds.
        name: Name of the command.
        options: Arguments/options the command was run with.
        is_executed_from_command_line: Flag that indicates if command was run from the command line.
        output: Standard and error output of the command.
        is_successful: Flag that indicates if command finished successfully.
    """
    name = models.CharField(
        max_length=250,
        blank=False,
        null=False,
        editable=False,
        db_index=True
    )
    input = models.TextField(
        blank=False,
        null=False,
        editable=False
    )
    is_executed_from_command_line = models.BooleanField(
        blank=False,
        null=False,
        default=False,
        editable=False
    )
    output = models.TextField(
        blank=True,
        null=True,
        editable=False
    )
    state = IntegerEnumField(
        null=False,
        blank=False,
        enum=CommandState,
        default=CommandState.ACTIVE,
        db_index=True
    )
    related_objects = MultipleDBGenericManyToManyField()

    objects = BaseLogQuerySet.as_manager()

    class Meta:
        ordering = ('-start',)


class CeleryTaskInvocationLogManager(models.Manager):

    def filter_processing(self, related_objects=None, **kwargs):
        qs = self.filter(
            state=CeleryTaskInvocationLogState.TRIGGERED
        ).filter(**kwargs)

        if related_objects:
            for related_object in related_objects:
                qs = qs.filter(
                    related_objects__object_id=related_object.pk,
                    related_objects__object_ct_id=ContentType.objects.get_for_model(related_object).pk
                )
        return qs


class CeleryTaskInvocationLog(CeleryTaskInvocationLogStrMixin, Log):

    celery_task_id = models.UUIDField(
        max_length=250,
        db_index=True,
        null=True,
        blank=True
    )
    name = models.CharField(
        null=False,
        blank=False,
        max_length=250,
        db_index=True
    )
    queue_name = models.CharField(
        null=True,
        blank=True,
        max_length=250
    )
    applied_at = models.DateTimeField(
        null=False,
        blank=False
    )
    triggered_at = models.DateTimeField(
        null=True,
        blank=True
    )
    is_unique = models.BooleanField()
    is_async = models.BooleanField()
    is_on_commit = models.BooleanField()
    input = models.TextField(
        blank=True,
        null=True,
        editable=False
    )
    task_args = models.JSONField(
        null=True,
        blank=True,
        editable=False,
        encoder=DjangoJSONEncoder
    )
    task_kwargs = models.JSONField(
        null=True,
        blank=True,
        editable=False,
        encoder=DjangoJSONEncoder
    )
    estimated_time_of_first_arrival = models.DateTimeField(
        null=True,
        blank=True
    )
    expires_at = models.DateTimeField(
        null=True,
        blank=True
    )
    stale_at = models.DateTimeField(
        null=True,
        blank=True
    )
    state = IntegerEnumField(
        null=False,
        blank=False,
        enum=CeleryTaskInvocationLogState,
        default=CeleryTaskInvocationLogState.WAITING,
        db_index=True
    )

    related_objects = MultipleDBGenericManyToManyField()

    objects = CeleryTaskInvocationLogManager.from_queryset(BaseLogQuerySet)()

    class Meta:
        ordering = ('-start',)

    def __str__(self):
        return '{} ({})'.format(self.name, self.state.label)

    @property
    def runs(self):
        return CeleryTaskRunLog.objects.filter(celery_task_id=self.celery_task_id)

    @property
    def last_run(self):
        return CeleryTaskRunLog.objects.filter(celery_task_id=self.celery_task_id).first()

    @property
    def first_run(self):
        return CeleryTaskRunLog.objects.filter(celery_task_id=self.celery_task_id).last()


class CeleryTaskRunLog(CeleryTaskRunLogStrMixin, Log):

    celery_task_id = models.UUIDField(
        max_length=250,
        db_index=True,
        null=True,
        blank=True
    )
    name = models.CharField(
        null=False,
        blank=False,
        max_length=250,
        db_index=True
    )
    queue_name = models.CharField(
        null=True,
        blank=True,
        max_length=250
    )
    input = models.TextField(
        blank=True,
        null=True,
        editable=False
    )
    task_args = models.JSONField(
        null=True,
        blank=True,
        editable=False,
        encoder=DjangoJSONEncoder
    )
    task_kwargs = models.JSONField(
        null=True,
        blank=True,
        editable=False,
        encoder=DjangoJSONEncoder
    )
    state = IntegerEnumField(
        null=False,
        blank=False,
        enum=CeleryTaskRunLogState,
        default=CeleryTaskRunLogState.ACTIVE,
        db_index=True
    )
    result = models.JSONField(
        blank=True,
        null=True,
        editable=False,
        encoder=DjangoJSONEncoder
    )
    output = models.TextField(
        blank=True,
        null=True,
        editable=False
    )
    retries = models.PositiveSmallIntegerField(
        null=False,
        blank=False,
        default=0
    )
    estimated_time_of_next_retry = models.DateTimeField(
        null=True,
        blank=True
    )
    waiting_time = models.FloatField(
        null=True,
        blank=True
    )
    related_objects = MultipleDBGenericManyToManyField()

    objects = BaseLogQuerySet.as_manager()

    class Meta:
        ordering = ('-start',)

    def __str__(self):
        return '{} ({})'.format(self.name, self.get_state_display())

    def get_task_invocation_logs(self):
        return CeleryTaskInvocationLog.objects.filter(celery_task_id=self.celery_task_id)


def get_response_state(status_code):
    if status_code >= 500:
        return RequestLogState.ERROR
    elif status_code >= 400:
        return RequestLogState.WARNING
    else:
        return RequestLogState.INFO


logger_name_to_log_model = {
    LoggerName.INPUT_REQUEST: InputRequestLog,
    LoggerName.OUTPUT_REQUEST: OutputRequestLog,
    LoggerName.COMMAND: CommandLog,
    LoggerName.CELERY_TASK_INVOCATION: CeleryTaskInvocationLog,
    LoggerName.CELERY_TASK_RUN: CeleryTaskRunLog,
}


def get_log_model_from_logger_name(logger_name):
    return logger_name_to_log_model[logger_name]


def get_log_from_key(key):
    try:
        logger_name, id = key.split('|')
        return get_log_model_from_logger_name(logger_name).get(id=id)
    except ObjectDoesNotExist:
        return None


def get_log_key(log):
    logger_name = {v: k for k, v in logger_name_to_log_model.items()}[log.__class__]
    return '{}|{}'.format(logger_name.value, log.id)



from django.core.management.base import BaseCommand

from security.backends.writer import set_stale_celery_task_log_state


class Command(BaseCommand):

    def handle(self, **options):
        set_stale_celery_task_log_state()



from django.utils.timezone import now

from security.enums import LoggerName
from security.logging.common import SecurityLogger
from security.backends.signals import (
    command_started, command_output_updated, command_finished, command_error,
)


class CommandLogger(SecurityLogger):

    logger_name = LoggerName.COMMAND

    def __init__(self, name=None, input=None, is_executed_from_command_line=None, output=None, **kwargs):
        super().__init__(**kwargs)
        self.name = name
        self.input = input
        self.is_executed_from_command_line = is_executed_from_command_line
        self.output = output

    def log_started(self, name, input, is_executed_from_command_line):
        self.start = now()
        self.name = name
        self.input = input
        self.is_executed_from_command_line = is_executed_from_command_line
        command_started.send(sender=CommandLogger, logger=self)

    def log_finished(self):
        self.stop = now()
        command_finished.send(sender=CommandLogger, logger=self)

    def log_output_updated(self, output):
        self.output = output
        command_output_updated.send(sender=CommandLogger, logger=self)

    def log_exception(self, ex_tb):
        self.stop = now()
        self.error_message = ex_tb,
        command_error.send(sender=CommandLogger, logger=self)



import functools
import traceback

import requests

from io import BytesIO

from suds.client import Client as DefaultClient
from suds.transport import Reply, TransportError
from suds.transport.http import HttpTransport

import requests as security_requests


def handle_errors(f):
    @functools.wraps(f)
    def wrapper(*args, **kwargs):
        try:
            return f(*args, **kwargs)
        except requests.HTTPError as e:
            buf = BytesIO(e.response.content)
            raise TransportError(
                'Error in requests\n' + traceback.format_exc(), e.response.status_code, buf,
            )
        except requests.RequestException:
            raise TransportError(
                'Error in requests\n' + traceback.format_exc(), 000,
            )
    return wrapper


class SecurityRequestsTransport(HttpTransport):

    def __init__(self, slug=None, session=None, related_objects=None, timeout=None):
        self.related_objects = related_objects or ()
        self.slug = slug
        self.timeout = timeout
        # super won't work because not using new style class
        HttpTransport.__init__(self)
        self._session = session or security_requests.SecuritySession()

    @handle_errors
    def open(self, request):
        url = request.url
        if url.startswith('file:'):
            return HttpTransport.open(self, request)
        else:
            resp = self._session.get(url)
            resp.raise_for_status()
            return BytesIO(resp.content)

    @handle_errors
    def send(self, request):
        try:
            resp = self._session.post(request.url, data=request.message, headers=request.headers, slug=self.slug,
                                      related_objects=self.related_objects, timeout=self.timeout)
            if resp.headers.get('content-type') not in {'text/xml', 'application/soap+xml'}:
                resp.raise_for_status()
            return Reply(resp.status_code, resp.headers, resp.content)
        finally:
            self.related_objects = ()

    def add_related_objects(self, *related_objects):
        self.related_objects += tuple(related_objects)


class Client(DefaultClient):

    def __init__(self, url, slug=None, related_objects=None, session=None, transport=None, timeout=None, **kwargs):
        transport = transport or SecurityRequestsTransport(slug=slug, related_objects=related_objects, session=session,
                                                           timeout=timeout)
        DefaultClient.__init__(self, url, transport=transport, **kwargs)

    def add_related_objects(self, *related_objects):
        self.options.transport.add_related_objects(*related_objects)



from datetime import timedelta

from django.utils import timezone
from django.utils.translation import ugettext_lazy as _
from django.urls import resolve, Resolver404

from ipware.ip import get_client_ip

from security.config import settings
from security.backends.reader import get_count_input_requests
from security.enums import InputRequestSlug

from .exception import ThrottlingException


class ThrottlingValidator:

    def __init__(self, timeframe, throttle_at, description):
        self.timeframe = timeframe
        self.throttle_at = throttle_at
        self.description = description

    def validate(self, request):
        if settings.THROTTLING_ENABLED and not self._validate(request):
            raise ThrottlingException(self.description)

    def _validate(self, request):
        raise NotImplementedError

    def __repr__(self):
        return '<{} (timeframe={}, throttle_at={}, description={})>'.format(
            self.__class__.__name__, self.timeframe, self.throttle_at, self.description
        )


class PerRequestThrottlingValidator(ThrottlingValidator):

    def __init__(self, timeframe, throttle_at, description=_('Slow down')):
        super().__init__(timeframe, throttle_at, description)

    def _validate(self, request):
        try:
            view_slug = resolve(request.path_info, getattr(request, 'urlconf', None)).view_name
        except Resolver404:
            view_slug = None

        current_logger = getattr(request, 'input_request_logger', None)
        return get_count_input_requests(
            from_time=timezone.now() - timedelta(seconds=self.timeframe),
            ip=get_client_ip(request)[0],
            path=request.path,
            method=request.method.upper(),
            view_slug=view_slug,
            exclude_log_id=current_logger.id if current_logger else None
        ) < self.throttle_at


class LoginThrottlingValidator(ThrottlingValidator):

    slug = None

    def _validate(self, request):
        current_logger = getattr(request, 'input_request_logger', None)
        return get_count_input_requests(
            from_time=timezone.now() - timedelta(seconds=self.timeframe),
            ip=get_client_ip(request)[0],
            path=request.path,
            slug=self.slug,
            exclude_log_id=current_logger.id if current_logger else None
        ) < self.throttle_at


class UnsuccessfulLoginThrottlingValidator(LoginThrottlingValidator):

    slug = InputRequestSlug.UNSUCCESSFUL_LOGIN_REQUEST

    def __init__(self, timeframe, throttle_at, description=_('Too many login attempts')):
        super().__init__(timeframe, throttle_at, description)


class SuccessfulLoginThrottlingValidator(LoginThrottlingValidator):

    slug = InputRequestSlug.SUCCESSFUL_LOGIN_REQUEST

    def __init__(self, timeframe, throttle_at, description=_('You have logged in too many times')):
        super().__init__(timeframe, throttle_at, description)


class UnSuccessfulTwoFactorCodeVerificationThrottlingValidator(LoginThrottlingValidator):

    slug = InputRequestSlug.UNSUCCESSFUL_2FA_CODE_VERIFICATION_REQUEST

    def __init__(self, timeframe, throttle_at, description=_('Too many login attempts')):
        super().__init__(timeframe, throttle_at, description)


class SuccessfulTwoFactorCodeVerificationThrottlingValidator(LoginThrottlingValidator):

    slug = InputRequestSlug.SUCCESSFUL_2FA_CODE_VERIFICATION_REQUEST

    def __init__(self, timeframe, throttle_at, description=_('You have logged in too many times')):
        super().__init__(timeframe, throttle_at, description)



import datetime
import uuid
import decimal

from django.utils.timezone import now

from germanium.test_cases.default import GermaniumTestCase
from germanium.tools import assert_equal
from germanium.decorators import data_consumer

from freezegun import freeze_time

from django.test import override_settings

from security.utils import serialize_data, deserialize_data, truncate_lines_from_left, LogStringIO


class UtilsTestCase(GermaniumTestCase):

    def test_utils_serialize_and_deserialize_data_should_return_right_values(self):
        data = {
            'a': datetime.datetime(2018, 1, 10, 5, 47, 11),
            'b': [datetime.date(2020, 5, 21), datetime.time(16, 59)],
            'c': {
                'd': uuid.UUID('06335e84-2872-4914-8c5d-3ed07d2a2f16'),
                'e': decimal.Decimal('158.687')
            },
            'f': datetime.timedelta(days=4, hours=15, minutes=6, seconds=45),
            'g': [5, 'test'],
        }
        serialized_data = {
            'a': {'@type': 'datetime', '@value': '2018-01-10T05:47:11'},
            'b': [{'@type': 'date', '@value': '2020-05-21'}, {'@type': 'time', '@value': '16:59:00'}],
            'c': {
                'd': {'@type': 'uuid', '@value': '06335e84-2872-4914-8c5d-3ed07d2a2f16'},
                'e': {'@type': 'decimal', '@value': '158.687'}
            },
            'f': {'@type': 'timedelta', '@value': 'P4DT15H06M45S'},
            'g': [5, 'test']
        }
        assert_equal(serialize_data(data), serialized_data)
        assert_equal(deserialize_data(serialized_data), data)

    @data_consumer(
        (
            ('test', 'test', 10),
            ('long truncated value without newline', '…\n', 10),
            ('long truncated value with\nnewline', '…\nnewline', 10),
            ('long truncated value with\nnewline', '…\n', 5),
            ('long truncated value with newline after end \n', '…\n', 5)
        )
    )
    def test_truncate_lines_from_left_should_truncate_text(self, original_text, truncated_text, length):
        assert_equal(truncate_lines_from_left(original_text, length), truncated_text)

    @override_settings(
        SECURITY_LOG_STRING_OUTPUT_TRUNCATE_LENGTH=100,
        SECURITY_LOG_STRING_OUTPUT_TRUNCATE_OFFSET=10
    )
    def test_log_string_io_should_be_truncated_according_to_settings(self):
        log_string_io = LogStringIO(write_time=False)
        for _ in range(20):
            log_string_io.write((4*'a') + '\n')
        assert_equal(log_string_io.getvalue(), 20*(4*'a' + '\n'))
        assert_equal(len(log_string_io.getvalue()), 100)
        log_string_io.write('b')
        assert_equal(log_string_io.getvalue(), '…\n' + 17*(4*'a' + '\n') + 'b')

    @override_settings(
        SECURITY_LOG_STRING_OUTPUT_TRUNCATE_LENGTH=100,
        SECURITY_LOG_STRING_OUTPUT_TRUNCATE_OFFSET=10
    )
    def test_log_string_io_should_remove_line_with_cr_lf(self):
        log_string_io = LogStringIO(write_time=False)
        log_string_io.write('text before newline\n')
        log_string_io.write('removed text')
        assert_equal(log_string_io.getvalue(), 'text before newline\nremoved text')
        log_string_io.write('\rnew text')
        assert_equal(log_string_io.getvalue(), 'text before newline\nnew text')

    @override_settings(
        SECURITY_LOG_STRING_OUTPUT_TRUNCATE_LENGTH=100,
        SECURITY_LOG_STRING_OUTPUT_TRUNCATE_OFFSET=10
    )
    def test_log_string_io_should_truncate_and_remove_last_line(self):
        log_string_io = LogStringIO(write_time=False)
        log_string_io.write('very long text which should be removed\n')
        log_string_io.write('very long text which shouldn\'t be removed\n')
        assert_equal(
            log_string_io.getvalue(),
            'very long text which should be removed\n'
            'very long text which shouldn\'t be removed\n'
        )
        log_string_io.write('text which should be removed with cr lf')
        assert_equal(
            log_string_io.getvalue(),
            '…\n'
            'very long text which shouldn\'t be removed\n'
            'text which should be removed with cr lf'
        )

        log_string_io.write('\rnew text')
        assert_equal(
            log_string_io.getvalue(),
            '…\n'
            'very long text which shouldn\'t be removed\n'
            'new text'
        )

        log_string_io.write('\rnew very long text which should remove prev text' + 9*' ')
        assert_equal(
            log_string_io.getvalue(),
            '…\n'
            'new very long text which should remove prev text' + 9*' '
        )

        log_string_io.write('\rnew very long text which should be removed at all' + 50*' ')
        assert_equal(
            log_string_io.getvalue(),
            '…\n'
        )

    @freeze_time(now())
    def test_log_string_io_should_add_time_and_log_to_output(self):
        time_string = f'\x1b[0m\x1b[1m[{{}}] [{now().strftime("%d-%m-%Y %H:%M:%S")}]\x1b[0m '

        log_string_io = LogStringIO()
        log_string_io.write('test')
        assert_equal(log_string_io.getvalue(), time_string.format(1) + 'test')

        log_string_io.write(' test')
        assert_equal(log_string_io.getvalue(), time_string.format(1) + 'test test')

        log_string_io.write('\ntest')
        assert_equal(
            log_string_io.getvalue(),
            time_string.format(1) + 'test test\n'
            + time_string.format(2) + 'test'
        )

    @override_settings(
        SECURITY_LOG_STRING_OUTPUT_TRUNCATE_LENGTH=110,
        SECURITY_LOG_STRING_OUTPUT_TRUNCATE_OFFSET=10
    )
    @freeze_time(now())
    def test_log_string_io_should_add_time_and_log_to_output_but_remove_with_cr_lf(self):
        time_string = f'\x1b[0m\x1b[1m[{{}}] [{now().strftime("%d-%m-%Y %H:%M:%S")}]\x1b[0m '

        log_string_io = LogStringIO()
        log_string_io.write('long text that should be removed \n')
        assert_equal(
            log_string_io.getvalue(),
            time_string.format(1) + 'long text that should be removed \n' + time_string.format(2)
        )

        log_string_io.write('new text')
        assert_equal(
            log_string_io.getvalue(),
            '…\n'
            + time_string.format(2) + 'new text'
        )

        log_string_io.write('write very long text' + 100*' ')
        assert_equal(
            log_string_io.getvalue(),
            '…\n' + time_string.format(2)
        )

        log_string_io.write('new text')
        assert_equal(
            log_string_io.getvalue(),
            '…\n' + time_string.format(2) + 'new text'
        )



# Generated by Django 3.1.14 on 2022-03-11 15:11

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('security_backends_sql', '0003_migration'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='celerytaskinvocationlog',
            name='is_duplicate',
        ),
        migrations.AddField(
            model_name='celerytaskinvocationlog',
            name='error_message',
            field=models.TextField(blank=True, null=True),
        ),
        migrations.AlterField(
            model_name='celerytaskinvocationloggenericmanytomanyrelation',
            name='changed_at',
            field=models.DateTimeField(auto_now=True, db_index=True, verbose_name='changed at'),
        ),
        migrations.AlterField(
            model_name='celerytaskinvocationloggenericmanytomanyrelation',
            name='created_at',
            field=models.DateTimeField(auto_now_add=True, db_index=True, verbose_name='created at'),
        ),
        migrations.AlterField(
            model_name='celerytaskinvocationloggenericmanytomanyrelation',
            name='id',
            field=models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID'),
        ),
        migrations.AlterField(
            model_name='celerytaskinvocationloggenericmanytomanyrelation',
            name='object_ct_id',
            field=models.PositiveSmallIntegerField(db_index=True, verbose_name='content type of the related object'),
        ),
        migrations.AlterField(
            model_name='celerytaskinvocationloggenericmanytomanyrelation',
            name='object_id',
            field=models.TextField(db_index=True, verbose_name='ID of the related object'),
        ),
        migrations.AlterField(
            model_name='celerytaskrunloggenericmanytomanyrelation',
            name='changed_at',
            field=models.DateTimeField(auto_now=True, db_index=True, verbose_name='changed at'),
        ),
        migrations.AlterField(
            model_name='celerytaskrunloggenericmanytomanyrelation',
            name='created_at',
            field=models.DateTimeField(auto_now_add=True, db_index=True, verbose_name='created at'),
        ),
        migrations.AlterField(
            model_name='celerytaskrunloggenericmanytomanyrelation',
            name='id',
            field=models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID'),
        ),
        migrations.AlterField(
            model_name='celerytaskrunloggenericmanytomanyrelation',
            name='object_ct_id',
            field=models.PositiveSmallIntegerField(db_index=True, verbose_name='content type of the related object'),
        ),
        migrations.AlterField(
            model_name='celerytaskrunloggenericmanytomanyrelation',
            name='object_id',
            field=models.TextField(db_index=True, verbose_name='ID of the related object'),
        ),
        migrations.AlterField(
            model_name='commandlog',
            name='error_message',
            field=models.TextField(blank=True, null=True),
        ),
        migrations.AlterField(
            model_name='commandloggenericmanytomanyrelation',
            name='changed_at',
            field=models.DateTimeField(auto_now=True, db_index=True, verbose_name='changed at'),
        ),
        migrations.AlterField(
            model_name='commandloggenericmanytomanyrelation',
            name='created_at',
            field=models.DateTimeField(auto_now_add=True, db_index=True, verbose_name='created at'),
        ),
        migrations.AlterField(
            model_name='commandloggenericmanytomanyrelation',
            name='id',
            field=models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID'),
        ),
        migrations.AlterField(
            model_name='commandloggenericmanytomanyrelation',
            name='object_ct_id',
            field=models.PositiveSmallIntegerField(db_index=True, verbose_name='content type of the related object'),
        ),
        migrations.AlterField(
            model_name='commandloggenericmanytomanyrelation',
            name='object_id',
            field=models.TextField(db_index=True, verbose_name='ID of the related object'),
        ),
        migrations.AlterField(
            model_name='inputrequestloggenericmanytomanyrelation',
            name='changed_at',
            field=models.DateTimeField(auto_now=True, db_index=True, verbose_name='changed at'),
        ),
        migrations.AlterField(
            model_name='inputrequestloggenericmanytomanyrelation',
            name='created_at',
            field=models.DateTimeField(auto_now_add=True, db_index=True, verbose_name='created at'),
        ),
        migrations.AlterField(
            model_name='inputrequestloggenericmanytomanyrelation',
            name='id',
            field=models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID'),
        ),
        migrations.AlterField(
            model_name='inputrequestloggenericmanytomanyrelation',
            name='object_ct_id',
            field=models.PositiveSmallIntegerField(db_index=True, verbose_name='content type of the related object'),
        ),
        migrations.AlterField(
            model_name='inputrequestloggenericmanytomanyrelation',
            name='object_id',
            field=models.TextField(db_index=True, verbose_name='ID of the related object'),
        ),
        migrations.AlterField(
            model_name='outputrequestloggenericmanytomanyrelation',
            name='changed_at',
            field=models.DateTimeField(auto_now=True, db_index=True, verbose_name='changed at'),
        ),
        migrations.AlterField(
            model_name='outputrequestloggenericmanytomanyrelation',
            name='created_at',
            field=models.DateTimeField(auto_now_add=True, db_index=True, verbose_name='created at'),
        ),
        migrations.AlterField(
            model_name='outputrequestloggenericmanytomanyrelation',
            name='id',
            field=models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID'),
        ),
        migrations.AlterField(
            model_name='outputrequestloggenericmanytomanyrelation',
            name='object_ct_id',
            field=models.PositiveSmallIntegerField(db_index=True, verbose_name='content type of the related object'),
        ),
        migrations.AlterField(
            model_name='outputrequestloggenericmanytomanyrelation',
            name='object_id',
            field=models.TextField(db_index=True, verbose_name='ID of the related object'),
        ),
        migrations.AddField(
            model_name='celerytaskinvocationlog',
            name='version',
            field=models.PositiveSmallIntegerField(default=9999),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='celerytaskrunlog',
            name='version',
            field=models.PositiveSmallIntegerField(default=9999),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='commandlog',
            name='version',
            field=models.PositiveSmallIntegerField(default=9999),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='inputrequestlog',
            name='version',
            field=models.PositiveSmallIntegerField(default=9999),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='outputrequestlog',
            name='version',
            field=models.PositiveSmallIntegerField(default=9999),
            preserve_default=False,
        ),
    ]



import os
import json
import gzip
import logging
import time as time_sleep

from itertools import islice, chain

from enum import Enum

from io import TextIOWrapper

from datetime import datetime, time, timedelta

from elasticsearch_dsl.utils import AttrDict, AttrList

from django.utils.timezone import now, utc
from django.utils.module_loading import import_string
from django.core.serializers.json import DjangoJSONEncoder

from celery import current_app
from celery.exceptions import NotRegistered

from elasticsearch import ConflictError, ConnectionTimeout
from elasticsearch_dsl import Q

from security.config import settings
from security.enums import (
    RequestLogState, CeleryTaskInvocationLogState, CeleryTaskRunLogState, CommandState
)
from security.backends.writer import BaseBackendWriter

from .models import (
    CeleryTaskInvocationLog, get_key_from_content_type_object_id_and_model_db, get_response_state,
    get_log_model_from_logger_name, get_object_triple_from_key, logger_name_to_log_model, get_index_name
)


logstash_logger = logging.getLogger('security.logstash')

MAX_VERSION = 9999


def batch_delete_queryset(queryset, batch, stdout, stderr, attempt=0):
    try:
        tmp_params = queryset._params.copy()
        queryset._params['max_docs'] = batch
        queryset._params['refresh'] = True
        queryset._params['request_timeout'] = settings.ELASTICSEARCH_CLEAN_LOGS_DELETE_REQUEST_TIMEOUT
        while True:
            stdout.write('.', ending='')
            if queryset.delete()['deleted'] == 0:
                break
        queryset._params = tmp_params
    except (ConflictError, ConnectionTimeout):
        if attempt > settings.ELASTICSEARCH_CLEAN_LOGS_DELETE_ATTEMPTS:
            raise
        time_sleep.sleep(settings.ELASTICSEARCH_CLEAN_LOGS_DELETE_ATTEMPT_SLEEP)
        stderr.write('.', ending='')
        queryset._params = tmp_params
        batch_delete_queryset(queryset, batch, stdout, stderr, attempt+1)


def batch(iterable, size):
    sourceiter = iter(iterable)
    try:
        while True:
            batchiter = islice(sourceiter, size)
            yield chain([next(batchiter)], batchiter)
    except StopIteration:
        pass


def get_queryset_dict_data_by_batch(qs, batch_size):
    for batch_qs in get_querysets_by_batch(qs, batch_size):
        for obj in batch_qs:
            yield obj.to_dict()


def get_lazy_serialize_iterable(iterable, stdout, batch):

    class StreamList(list):

        def __iter__(self):
            for i, obj in enumerate(iterable, 1):
                if i % batch == 0:
                    stdout.write('.', ending='')
                yield obj

        def __len__(self):
            return 1

    return StreamList()


def get_querysets_by_batch(qs, batch):
    last_document = None
    while True:
        batch_qs = qs
        if last_document:
            batch_qs = batch_qs.extra(search_after=last_document.meta.sort)

        batch_list = list(batch_qs[:batch])
        yield batch_list

        if len(batch_list) == batch:
            last_document = batch_list[-1]
        else:
            break


class BaseElasticsearchDataWriter:

    def _get_related_object_keys(self, logger):
        return [
            get_key_from_content_type_object_id_and_model_db(content_type_pk, object_pk, model_db)
            for model_db, content_type_pk, object_pk in logger.related_objects
        ]

    def _get_index_data(self, index):
        index_data = {}
        for field_name in index._doc_type.mapping.properties.to_dict()['properties'].keys():
            value = getattr(index, field_name)
            if isinstance(value, AttrList):
                value = value._l_
            elif isinstance(value, AttrDict):
                value = value.to_dict()

            index_data[field_name] = value
        return index_data

    def update_index(self, index, **updated_data):
        raise NotImplementedError

    def create_or_update_index_from_logger(self, logger, is_last=False, **extra_data):
        raise NotImplementedError


class DirectElasticsearchDataWriter(BaseElasticsearchDataWriter):

    def _create_index(self, logger_id, logger_name, version, data):
        try:
            index_class = get_log_model_from_logger_name(logger_name)
            index = index_class(**data)
            index.meta.id = logger_id
            index.save(
                params={
                    'version': version,
                    'version_type': 'external'
                }
            )
        except ConflictError:
            pass

    def update_index(self, index, **updated_data):
        data = self._get_index_data(index)
        data.update(updated_data)
        logger_name = {v: k for k, v in logger_name_to_log_model.items()}[index.__class__]
        self._create_index(
            index.id,
            logger_name,
            MAX_VERSION,
            data
        )

    def create_or_update_index_from_logger(self, logger, is_last=False, **extra_data):
        if is_last:
            version = MAX_VERSION
        elif 'last_elasticsearch_version' in logger.backend_logs:
            version = logger.backend_logs['last_elasticsearch_version'] + 1
        else:
            version = 0

        data = logger.to_dict()

        data['related_objects'] = self._get_related_object_keys(logger)
        data.update(extra_data)

        self._create_index(
            logger.id,
            logger.logger_name,
            version,
            data
        )
        logger.backend_logs['last_elasticsearch_version'] = version


class LogstashElasticsearchDataWriter(BaseElasticsearchDataWriter):

    def _serialize_data(self, data):
        serialized_data = {}
        for k, v in data.items():
            if isinstance(v, Enum):
                serialized_data[k] = v.name
            elif isinstance(v, (list, dict, tuple)) and k not in {'extra_data', 'related_objects'}:
                serialized_data[k] = json.dumps(v, cls=DjangoJSONEncoder)
            else:
                serialized_data[k] = v
        return json.dumps(serialized_data, cls=DjangoJSONEncoder)

    def _get_log_message(self, logger_id, logger_name, version, data):
        return f'{get_index_name(logger_name)} {version} {logger_id} {self._serialize_data(data)}'

    def _get_logger_message(self, logger, last_version=False, **extra_data):
        if last_version:
            version = MAX_VERSION
        elif 'last_elasticsearch_version' in logger.backend_logs:
            version = logger.backend_logs['last_elasticsearch_version'] + 1
        else:
            version = 0

        logger_data = logger.to_dict()
        logger_data['related_objects'] = self._get_related_object_keys(logger)
        logger_data.update(extra_data)

        logger.backend_logs['last_elasticsearch_version'] = version
        return self._get_log_message(
            logger.id,
            logger.logger_name,
            version,
            logger_data
        )

    def create_or_update_index_from_logger(self, logger, is_last=False, **extra_data):
        logstash_logger.info(
            self._get_logger_message(
                logger,
                last_version=is_last,
                **extra_data
            )
        )

    def update_index(self, index, **updated_data):
        data = self._get_index_data(index)
        data.update(updated_data)
        logger_name = {v: k for k, v in logger_name_to_log_model.items()}[index.__class__]
        logstash_logger.info(
            self._get_log_message(
                index.id,
                logger_name,
                MAX_VERSION,
                data
            )
        )


class ElasticsearchBackendWriter(BaseBackendWriter):

    def get_data_writer(self):
        return (
            LogstashElasticsearchDataWriter()
            if settings.ELASTICSEARCH_LOGSTASH_WRITER else DirectElasticsearchDataWriter()
        )

    def input_request_started(self, logger):
        self.get_data_writer().create_or_update_index_from_logger(logger, state=RequestLogState.INCOMPLETE)

    def input_request_finished(self, logger):
        self.get_data_writer().create_or_update_index_from_logger(
            logger, is_last=True, state=get_response_state(logger.response_code)
        )

    def input_request_error(self, logger):
        self.get_data_writer().create_or_update_index_from_logger(logger, state=RequestLogState.ERROR)

    def output_request_started(self, logger):
        self.get_data_writer().create_or_update_index_from_logger(logger, state=RequestLogState.INCOMPLETE)

    def output_request_finished(self, logger):
        self.get_data_writer().create_or_update_index_from_logger(
            logger, is_last=True, state=get_response_state(logger.response_code)
        )

    def output_request_error(self, logger):
        self.get_data_writer().create_or_update_index_from_logger(
            logger, is_last=True, state=RequestLogState.ERROR
        )

    def command_started(self, logger):
        self.get_data_writer().create_or_update_index_from_logger(
            logger, state=CommandState.ACTIVE
        )

    def command_output_updated(self, logger):
        self.get_data_writer().create_or_update_index_from_logger(
            logger, state=CommandState.ACTIVE
        )

    def command_finished(self, logger):
        self.get_data_writer().create_or_update_index_from_logger(
            logger, is_last=True, state=CommandState.SUCCEEDED
        )

    def command_error(self, logger):
        self.get_data_writer().create_or_update_index_from_logger(
            logger, is_last=True, state=CommandState.FAILED
        )

    def celery_task_invocation_started(self, logger):
        self.get_data_writer().create_or_update_index_from_logger(
            logger, state=CeleryTaskInvocationLogState.WAITING
        )

    def celery_task_invocation_triggered(self, logger):
        self.get_data_writer().create_or_update_index_from_logger(
            logger, state=CeleryTaskInvocationLogState.TRIGGERED
        )

    def celery_task_invocation_duplicate(self, logger):
        self.get_data_writer().create_or_update_index_from_logger(
            logger, is_last=True, state=CeleryTaskInvocationLogState.DUPLICATE
        )

    def celery_task_invocation_ignored(self, logger):
        self.get_data_writer().create_or_update_index_from_logger(
            logger, is_last=True, state=CeleryTaskInvocationLogState.IGNORED
        )

    def celery_task_invocation_timeout(self, logger):
        self.get_data_writer().create_or_update_index_from_logger(
            logger, is_last=True, state=CeleryTaskInvocationLogState.TIMEOUT
        )

    def celery_task_invocation_expired(self, logger):
        self.get_data_writer().create_or_update_index_from_logger(
            logger, is_last=True, state=CeleryTaskInvocationLogState.EXPIRED
        )

    def celery_task_invocation_succeeded(self, logger):
        self.get_data_writer().create_or_update_index_from_logger(
            logger, is_last=True, state=CeleryTaskInvocationLogState.SUCCEEDED
        )

    def celery_task_invocation_failed(self, logger):
        self.get_data_writer().create_or_update_index_from_logger(
            logger, is_last=True, state=CeleryTaskInvocationLogState.FAILED
        )

    def celery_task_run_started(self, logger):
        self.get_data_writer().create_or_update_index_from_logger(
            logger, state=CeleryTaskRunLogState.ACTIVE
        )

    def celery_task_run_succeeded(self, logger):
        self.get_data_writer().create_or_update_index_from_logger(
            logger, is_last=True, state=CeleryTaskRunLogState.SUCCEEDED
        )

    def celery_task_run_failed(self, logger):
        self.get_data_writer().create_or_update_index_from_logger(
            logger, is_last=True, state=CeleryTaskRunLogState.FAILED
        )

    def celery_task_run_retried(self, logger):
        self.get_data_writer().create_or_update_index_from_logger(
            logger, is_last=True, state=CeleryTaskRunLogState.RETRIED
        )

    def celery_task_run_output_updated(self, logger):
        self.get_data_writer().create_or_update_index_from_logger(
            logger, state=CeleryTaskRunLogState.ACTIVE
        )

    def set_stale_celery_task_log_state(self):
        processsing_stale_tasks = CeleryTaskInvocationLog.search().filter(
            Q('range', stale_at={'lt': now()}) & Q('term', state=CeleryTaskInvocationLogState.TRIGGERED.name)
        ).sort('stale_at')
        for task in processsing_stale_tasks[:settings.SET_STALE_CELERY_INVOCATIONS_LIMIT_PER_RUN]:
            task_last_run = task.last_run
            if task_last_run and task_last_run.state == CeleryTaskRunLogState.SUCCEEDED:
                self.get_data_writer().update_index(
                    task,
                    state=CeleryTaskInvocationLogState.SUCCEEDED,
                    stop=task_last_run.stop,
                    time=(task_last_run.stop - task.start).total_seconds()
                )
            elif task_last_run and task_last_run.state == CeleryTaskRunLogState.FAILED:
                self.get_data_writer().update_index(
                    task,
                    state=CeleryTaskInvocationLogState.FAILED,
                    stop=task_last_run.stop,
                    time=(task_last_run.stop - task.start).total_seconds()
                )
            else:
                try:
                    task_args = json.loads(task.task_args)
                    task_kwargs = json.loads(task.task_kwargs)
                    current_app.tasks[task.name].expire_invocation(
                        task.id,
                        task_args,
                        task_kwargs,
                        dict(
                            slug=task.slug,
                            parent_log=task.parent_log,
                            related_objects=[
                                get_object_triple_from_key(key) for key in task.related_objects or ()
                            ],
                            start=task.start,
                            celery_task_id=task.celery_task_id,
                            stop=task.stop,
                            name=task.name,
                            queue_name=task.queue_name,
                            applied_at=task.applied_at,
                            triggered_at=task.triggered_at,
                            is_unique=task.is_unique,
                            is_async=task.is_async,
                            is_on_commit=task.is_on_commit,
                            input=task.input,
                            task_args=task_args,
                            task_kwargs=task_kwargs,
                            estimated_time_of_first_arrival=task.estimated_time_of_first_arrival,
                            expires_at=task.expires_at,
                            stale_at=task.stale_at,
                        )
                    )
                except NotRegistered:
                    self.get_data_writer().update_index(
                        task,
                        state=CeleryTaskInvocationLogState.EXPIRED,
                        stop=now()
                    )

    def clean_logs(self, type, timestamp, backup_path, stdout, stderr):
        storage = import_string(settings.BACKUP_STORAGE_CLASS)()

        qs = get_log_model_from_logger_name(type).search().filter(
            Q('range', start={'lt': timestamp})
        ).sort('start')
        step_timestamp = None
        if qs.count() != 0:
            step_timestamp = list(qs[0:1])[0].start

        while step_timestamp and step_timestamp < timestamp:
            min_timestamp = datetime.combine(step_timestamp, time.min).replace(tzinfo=utc)
            max_timestamp = datetime.combine(step_timestamp, time.max).replace(tzinfo=utc)

            qs_filtered_by_day = qs.filter(Q('range', start={'gte': min_timestamp, 'lte': max_timestamp})).sort(
                'start', 'id'
            )

            if qs_filtered_by_day.count() != 0:
                stdout.write(
                    2 * ' ' + 'Cleaning logs for date {} ({})'.format(
                        step_timestamp.date(), qs_filtered_by_day.count()
                    )
                )

                if backup_path:
                    created_log_files = []

                    try:
                        for batch_data in batch(
                                get_queryset_dict_data_by_batch(
                                    qs_filtered_by_day, settings.ELASTICSEARCH_CLEAN_LOGS_READ_BATCH_SIZE
                                ),
                                settings.CLEAN_LOGS_BACKUP_FILE_BATCH_SIZE):
                            log_file_name = os.path.join(backup_path, str(step_timestamp.date()))
                            if storage.exists(f'{log_file_name}.json.gz'):
                                i = 1
                                while storage.exists(f'{log_file_name}({i}).json.gz'):
                                    i += 1
                                log_file_name = f'{log_file_name}({i})'
                            log_file_name = f'{log_file_name}.json.gz'
                            stdout.write(4 * ' ' + f'generating backup file: {log_file_name}', ending=' ')
                            with storage.open(log_file_name, 'wb') as f:
                                with TextIOWrapper(gzip.GzipFile(filename='{}.json'.format(log_file_name),
                                                                 fileobj=f, mode='wb')) as gzf:
                                    created_log_files.append(log_file_name)
                                    json.dump(
                                        get_lazy_serialize_iterable(
                                            batch_data, stdout, settings.ELASTICSEARCH_CLEAN_LOGS_READ_BATCH_SIZE
                                        ),
                                        gzf, cls=DjangoJSONEncoder, indent=5
                                    )
                            stdout.write('')
                    except Exception:
                        stdout.write('')
                        stderr.write(4 * ' ' + 'raised exception removing created files')
                        for log_file_name in created_log_files:
                            storage.delete(log_file_name)
                        raise
                stdout.write(4 * ' ' + 'deleting logs', ending=' ')
                batch_delete_queryset(
                    qs_filtered_by_day, settings.ELASTICSEARCH_CLEAN_LOGS_DELETE_BATCH_SIZE, stdout, stderr
                )
                stdout.write('')

            step_timestamp = min_timestamp + timedelta(days=1)



from .celery_log import CeleryLogTestCase
from .command_log import CommandLogTestCase
from .input_request_log import InputRequestLogTestCase
from .output_request_log import OutputRequestLogTestCase
from .commands import CommandTestCase
from .utils import UtilsTestCase



import functools
import operator

from django.contrib.auth import get_user_model
from django.db.models import TextField, Q
from django.db.models.functions import Cast
from django.utils.translation import ugettext
from django.contrib.contenttypes.models import ContentType

from pyston.filters.django_filters import SimpleMethodFilter
from pyston.filters.utils import OperatorSlug
from pyston.filters.exceptions import FilterValueError


class UsernameUserFilter(SimpleMethodFilter):

    allowed_operators = (OperatorSlug.CONTAINS,)

    def clean_value(self, value, operator_slug, request):
        user_model = get_user_model()
        return list(
            user_model.objects.filter(
                **{'{}__contains'.format(user_model.USERNAME_FIELD): value}
            ).annotate(
                str_id=Cast('id', output_field=TextField())
            ).values_list('str_id', flat=True)
        )

    def get_filter_term(self, value, operator_slug, request):
        return {'user_id__in': value}


class RelatedObjectsFilter(SimpleMethodFilter):

    allowed_operators = (OperatorSlug.IN,)

    def clean_value(self, value, operator_slug, request):
        cleaned_values = []
        for v in value:
            try:
                content_type_id, object_id = v.split('|')
                cleaned_values.append((int(content_type_id), object_id))
            except (ValueError, ContentType.DoesNotExist):
                raise FilterValueError(ugettext('Invalid value.'))
        return cleaned_values

    def get_filter_term(self, value, operator_slug, request):
        return functools.reduce(operator.or_, (
            Q(
                related_objects__object_id=object_id,
                related_objects__object_ct_id=object_ct_id
            ) for object_ct_id, object_id in value
        ))



import socket
from datetime import datetime

from logging.handlers import SocketHandler

from .serializer import serialize_message


def format_timestamp(time):
    tstamp = datetime.utcfromtimestamp(time)
    return tstamp.strftime("%Y-%m-%dT%H:%M:%S") + ".%03d" % (tstamp.microsecond / 1000) + "Z"


class TCPLogstashHandler(SocketHandler):
    """
    Python logging handler for Logstash. Sends events over TCP.
    :param host: The host of the logstash server.
    :param port: The port of the logstash server (default 5959).
    """

    def __init__(self, host, port=5959):
        super().__init__(host, port)
        self._host = socket.gethostname()

    def makePickle(self, record):
        return serialize_message(
            self.formatter.format(record),
            {
                'logger': {
                    'host': self._host,
                    'path': record.pathname,
                    'timestamp': format_timestamp(record.created),
                    'level': record.levelname,
                    'logger_name': record.name,
                },
                **getattr(record, 'metadata', {}),
            }
        )



import json

from io import StringIO

from contextlib import contextmanager

from django.contrib.auth.models import User

from germanium.decorators import data_provider
from germanium.tools import assert_equal, assert_false, assert_is_not_none

from security.management import call_command


TRUNCATION_CHAR = '…'


class BaseTestCaseMixin:

    databases = ['default', 'security']

    @data_provider
    def create_user(self, username='test', email='test@localhost'):
        return User.objects._create_user(username, email, 'test', is_staff=True, is_superuser=True)



def test_call_command(*args, **kwargs):
    call_command(*args, **kwargs, stdout=StringIO(), stderr=StringIO())


def assert_equal_logstash(logstash_output, expected_index, expected_version, expected_logger_id, expected_data):
    prefix_and_index, version, logger_id, data = logstash_output.split(' ', 3)
    assert_equal(prefix_and_index, f'INFO:security.logstash:{expected_index}')
    assert_equal(version, str(expected_version))
    assert_equal(logger_id, str(expected_logger_id))
    parsed_data = json.loads(data)
    for k, v in expected_data.items():
        assert_equal(parsed_data.get(k), v, f'Invalid data "{k}" ({parsed_data.get(k)} != {v})')


def assert_equal_log_data(captured_log, expected_data):
    for k, v in expected_data.items():
        assert_equal(getattr(captured_log, k), v)



from security.backends.reader import BaseBackendReader

from .models import InputRequestLog, get_log_model_from_logger_name


class SQLBackendReader(BaseBackendReader):

    def get_count_input_requests(self, from_time, ip=None, path=None, view_slug=None, slug=None, method=None,
                                 exclude_log_id=None):
        qs = InputRequestLog.objects.filter(
            start__gte=from_time,
        )
        if ip is not None:
            qs = qs.filter(ip=ip)
        if path is not None:
            qs = qs.filter(path=path)
        if method is not None:
            qs = qs.filter(method=method)
        if view_slug is not None:
            qs = qs.filter(view_slug=view_slug)
        if slug is not None:
            qs = qs.filter(slug=slug)
        if exclude_log_id is not None:
            qs = qs.exclude(pk=exclude_log_id)
        return qs.count()

    def get_logs_related_with_object(self, logger_name, related_object):
        return list(get_log_model_from_logger_name(logger_name).objects.filter_related_with_object(related_object))



from security.backends.reader import BaseBackendReader
from security.utils import get_object_triple

from .writer import capture_security_logs


class TestingBackendReader(BaseBackendReader):

    def get_count_input_requests(self, from_time, ip=None, path=None, view_slug=None, slug=None, method=None,
                                 exclude_log_id=None):
        if not capture_security_logs.logged_data:
            return 0

        count_requests = 0
        for logger in capture_security_logs.logged_data.input_request:
            if ((exclude_log_id is None or logger.id != exclude_log_id)
                    and (method is None or logger.method == method)
                    and (ip is None or logger.ip == ip)
                    and (path is None or logger.path == path)
                    and (view_slug is None or logger.view_slug == view_slug)
                    and (slug is None or logger.slug == slug)
                    and (logger.start >= from_time)):
                count_requests += 1
        return count_requests

    def get_logs_related_with_object(self, logger_name, related_object):
        if not capture_security_logs.logged_data:
            return []

        return [
            logger for logger in getattr(capture_security_logs.logged_data, logger_name.replace('-', '_'))
            if get_object_triple(related_object) in logger.related_objects
        ]



from django.core.exceptions import ImproperlyConfigured

from security.config import settings

from .app import SecurityBackend


class BaseBackendReader:

    def get_count_input_requests(self, from_time, ip=None, path=None, view_slug=None, slug=None, method=None,
                                 exclude_log_id=None):
        raise NotImplementedError

    def get_logs_related_with_object(self, logger_name, related_object):
        raise NotImplementedError

    def get_stale_celery_task_invocation_logs(self):
        raise NotImplementedError


def get_reader_backend():
    if not SecurityBackend.registered_readers:
        raise ImproperlyConfigured('No registered backend reader was set')
    if settings.BACKEND_READER is not None:
        if settings.BACKEND_READER not in SecurityBackend.registered_readers:
            raise ImproperlyConfigured(f'Backend reader "{settings.BACKEND_READER}" is not registered')
        return SecurityBackend.registered_readers[settings.BACKEND_READER]
    else:
        return next(iter(SecurityBackend.registered_readers.values()))


def get_count_input_requests(from_time, ip=None, path=None, view_slug=None, slug=None, method=None,
                             exclude_log_id=None):
    return get_reader_backend().get_count_input_requests(
        from_time, ip, path, view_slug, slug, method, exclude_log_id
    )


def get_logs_related_with_object(logger_name, related_object):
    return get_reader_backend().get_logs_related_with_object(
        logger_name, related_object
    )



from django.http import HttpResponse

from security.throttling.validators import PerRequestThrottlingValidator
from security.logging.requests.utils import log_input_request_with_data
from security.decorators import hide_request_body, log_exempt, throttling_exempt, throttling
from security import requests


def proxy_view(request):
    return HttpResponse(
        requests.get(
            request.GET['url'],
            slug='proxy',
            related_objects=(request.user,) if request.user.is_authenticated else ()
        ).content
    )


@hide_request_body()
def hide_request_body_view(request):
    return HttpResponse()


@log_exempt()
def log_exempt_view(request):
    return HttpResponse()


@throttling_exempt()
def throttling_exempt_view(request):
    return HttpResponse()


@throttling(PerRequestThrottlingValidator(60, 1))
def extra_throttling_view(request):
    return HttpResponse()


def error_view(request):
    raise RuntimeError


def home_view(request):
    if request.user.is_authenticated:
        log_input_request_with_data(
            request, slug='user-home', extra_data={'user_pk': request.user.pk}, related_objects=[request.user]
        )
    return HttpResponse('home page response')



from django.template.defaultfilters import truncatechars
from django.utils.translation import gettext_lazy as _l


class LogShortIdMixin:

    @property
    def short_id(self):
        return truncatechars(str(self.id), 8)


class LogStrMixin:

    VERBOSE_NAME = None

    def __str__(self):
        return '{} {}'.format(self.VERBOSE_NAME, self.short_id)


class InputRequestLogStrMixin(LogStrMixin):

    VERBOSE_NAME = _l('input request log')


class OutputRequestLogStrMixin(LogStrMixin):

    VERBOSE_NAME = _l('input request log')


class CommandLogStrMixin(LogStrMixin):

    VERBOSE_NAME = _l('command log')


class CeleryTaskInvocationLogStrMixin(LogStrMixin):

    VERBOSE_NAME = _l('celery task invocation log')


class CeleryTaskRunLogStrMixin(LogStrMixin):

    VERBOSE_NAME = _l('celery task run log')



import logging

from security.backends.writer import BaseBackendWriter


input_request_logger = logging.getLogger('security.input_request')
output_request_logger = logging.getLogger('security.output_request')
command_logger = logging.getLogger('security.command')
celery_logger = logging.getLogger('security.celery')


class LoggingBackendWriter(BaseBackendWriter):

    def input_request_started(self, logger):
        input_request_logger.info(
            'Input request "%(id)s" to "%(host)s" with path "%(path)s" was started',
            dict(
                id=logger.id,
                host=logger.host,
                path=logger.path,
            ),
            extra=dict(
                id=logger.id,
                input_request_host=logger.host,
                input_request_path=logger.path,
                input_request_method=logger.method,
                input_request_view_slug=logger.view_slug,
                input_request_slug=logger.slug,
                input_request_is_secure=logger.is_secure,
                input_request_ip=logger.ip,
                input_request_release=logger.release
            )
        )

    def input_request_finished(self, logger):
        input_request_logger.info(
            'Input request "%(id)s" to "%(host)s" with path "%(path)s" was finished',
            dict(
                id=logger.id,
                host=logger.host,
                path=logger.path,
            ),
            extra=dict(
                id=logger.id,
                input_request_host=logger.host,
                input_request_path=logger.path,
                input_request_method=logger.method,
                input_request_view_slug=logger.view_slug,
                input_request_slug=logger.slug,
                input_request_is_secure=logger.is_secure,
                input_request_ip=logger.ip,
                input_request_response_code=logger.response_code,
                input_request_start=logger.start,
                input_request_stop=logger.stop,
                input_request_time=logger.time,
                input_request_release=logger.release
            )
        )

    def input_request_error(self, logger):
        input_request_logger.error(
            'Input request "%(id)s" to "%(host)s" with path "%(path)s" failed',
            dict(
                id=logger.id,
                host=logger.host,
                path=logger.path,
            ),
            extra=dict(
                id=logger.id,
                input_request_host=logger.host,
                input_request_path=logger.path,
                input_request_method=logger.method,
                input_request_view_slug=logger.view_slug,
                input_request_slug=logger.slug,
                input_request_is_secure=logger.is_secure,
                input_request_ip=logger.ip,
                input_request_start=logger.start,
                input_request_release=logger.release
            )
        )

    def output_request_started(self, logger):
        output_request_logger.info(
            'Output request "%(id)s" to "%(host)s" with path "%(path)s" was started',
            dict(
                id=logger.id,
                host=logger.host,
                path=logger.path,
            ),
            extra=dict(
                id=logger.id,
                output_request_host=logger.host,
                output_request_path=logger.path,
                output_request_method=logger.method,
                output_request_slug=logger.slug,
                output_request_is_secure=logger.is_secure,
                output_request_release=logger.release
            )
        )

    def output_request_finished(self, logger):
        output_request_logger.info(
            'Output request "%(id)s" to "%(host)s" with path "%(path)s" was successful',
            dict(
                id=logger.id,
                host=logger.host,
                path=logger.path,
            ),
            extra=dict(
                id=logger.id,
                output_request_host=logger.host,
                output_request_path=logger.path,
                output_request_method=logger.method,
                output_request_slug=logger.slug,
                output_request_is_secure=logger.is_secure,
                output_request_response_code=logger.response_code,
                output_request_start=logger.start,
                output_request_stop=logger.stop,
                output_request_time=logger.time,
                output_request_release=logger.release
            )
        )

    def output_request_error(self, logger):
        output_request_logger.error(
            'Output request "%(id)s" to "%(host)s" with path "%(path)s" failed',
            dict(
                id=logger.id,
                host=logger.host,
                path=logger.path,
            ),
            extra=dict(
                id=logger.id,
                output_request_host=logger.host,
                output_request_path=logger.path,
                output_request_method=logger.method,
                output_request_slug=logger.slug,
                output_request_is_secure=logger.is_secure,
                output_request_start=logger.start,
                output_request_stop=logger.stop,
                output_request_time=logger.time,
                output_request_release=logger.release
            )
        )

    def command_started(self, logger):
        command_logger.info(
            'Command "%(id)s" with name "%(name)s" was started',
            dict(
                id=logger.id,
                name=logger.name,
            ),
            extra=dict(
                id=logger.id,
                command_name=logger.name,
                command_is_executed_from_command_line=logger.is_executed_from_command_line,
                command_start=logger.start,
                command_release=logger.release
            )
        )

    def command_output_updated(self, logger):
        """Output is not stored into log"""

    def command_finished(self, logger):
        command_logger.info(
            'Command "%(id)s" with name "%(name)s" was successful',
            dict(
                id=logger.id,
                name=logger.name,
            ),
            extra=dict(
                id=logger.id,
                command_name=logger.name,
                command_is_executed_from_command_line=logger.is_executed_from_command_line,
                command_start=logger.start,
                command_stop=logger.stop,
                command_time=logger.time,
                command_is_error=False,
                command_release=logger.release
            )
        )

    def command_error(self, logger):
        command_logger.error(
            'Command "%(id)s" with name "%(name)s" failed',
            dict(
                id=logger.id,
                name=logger.name,
            ),
            extra=dict(
                id=logger.id,
                command_name=logger.name,
                command_is_executed_from_command_line=logger.is_executed_from_command_line,
                command_start=logger.start,
                command_stop=logger.stop,
                command_time=logger.time,
                command_is_error=True,
                command_release=logger.release
            )
        )

    def celery_task_invocation_started(self, logger):
        """Invocation started means nothing for logging backend"""

    def celery_task_invocation_duplicate(self, logger):
        celery_logger.info(
            'Celery task invocation "%(id)s" with celery id "%(celery_task_id)s" and name "%(name)s" is duplicate',
            dict(
                id=logger.id,
                celery_task_id=logger.celery_task_id,
                name=logger.name,
            ),
            extra=dict(
                id=logger.id,
                celery_task_invocation_task_id=logger.celery_task_id,
                celery_task_invocation_name=logger.name,
                celery_task_invocation_queue=logger.queue_name,
                celery_task_invocation_start=logger.start,
                celery_task_invocation_state='DUPLICATE',
                celery_task_invocation_release=logger.release
            ),
        )

    def celery_task_invocation_triggered(self, logger):
        celery_logger.info(
            'Celery task invocation "%(id)s" with celery id "%(celery_task_id)s" and name "%(name)s" was invoked',
            dict(
                id=logger.id,
                celery_task_id=logger.celery_task_id,
                name=logger.name,
            ),
            extra=dict(
                id=logger.id,
                celery_task_invocation_task_id=logger.celery_task_id,
                celery_task_invocation_name=logger.name,
                celery_task_invocation_queue=logger.queue_name,
                celery_task_invocation_start=logger.start,
                celery_task_invocation_state='TRIGGERED',
                celery_task_invocation_release=logger.release
            ),
        )

    def celery_task_invocation_ignored(self, logger):
        celery_logger.info(
            'Celery task invocation "%(id)s" with name "%(name)s" was ignored',
            dict(
                id=logger.id,
                name=logger.name,
            ),
            extra=dict(
                id=logger.id,
                celery_task_invocation_name=logger.name,
                celery_task_invocation_queue=logger.queue_name,
                celery_task_invocation_start=logger.start,
                celery_task_invocation_stop=logger.stop,
                celery_task_invocation_time=logger.time,
                celery_task_invocation_state='IGNORED',
                celery_task_invocation_release=logger.release
            ),
        )

    def celery_task_invocation_timeout(self, logger):
        celery_logger.warning(
            'Celery task "%(id)s" with celery id "%(celery_task_id)s" and name "%(name)s" caused a response timeout',
            dict(
                id=logger.id,
                celery_task_id=logger.celery_task_id,
                name=logger.name,
            ),
            extra=dict(
                id=logger.id,
                celery_task_invocation_task_id=logger.celery_task_id,
                celery_task_invocation_name=logger.name,
                celery_task_invocation_queue=logger.queue_name,
                celery_task_invocation_start=logger.start,
                celery_task_invocation_stop=logger.stop,
                celery_task_invocation_time=logger.time,
                celery_task_invocation_state='TIMEOUT',
                celery_task_invocation_release=logger.release
            ),
        )

    def celery_task_invocation_succeeded(self, logger):
        celery_logger.info(
            'Celery task invocation "%(id)s" with celery id "%(celery_task_id)s" and name "%(name)s" was successful',
            dict(
                id=logger.id,
                celery_task_id=logger.celery_task_id or '',
                name=logger.name,
            ),
            extra=dict(
                id=logger.id,
                celery_task_invocation_task_id=logger.celery_task_id or '',
                celery_task_invocation_name=logger.name,
                celery_task_invocation_queue=logger.queue_name,
                celery_task_invocation_start=logger.start,
                celery_task_invocation_stop=logger.stop,
                celery_task_invocation_time=logger.time,
                celery_task_invocation_state='SUCCEEDED',
                celery_task_invocation_release=logger.release
            ),
        )

    def celery_task_invocation_failed(self, logger):
        celery_logger.info(
            'Celery task invocation "%(id)s" with celery id "%(celery_task_id)s" and name "%(name)s" was failed',
            dict(
                id=logger.id,
                celery_task_id=logger.celery_task_id or '',
                name=logger.name,
            ),
            extra=dict(
                id=logger.id,
                celery_task_invocation_task_id=logger.celery_task_id or '',
                celery_task_invocation_name=logger.name,
                celery_task_invocation_queue=logger.queue_name,
                celery_task_invocation_start=logger.start,
                celery_task_invocation_stop=logger.stop,
                celery_task_invocation_time=logger.time,
                celery_task_invocation_state='FAILED',
                celery_task_invocation_release=logger.release
            ),
        )

    def celery_task_invocation_expired(self, logger):
        celery_logger.error(
            'Celery task invocation "%(id)s" with celery id "%(celery_task_id)s" and name "%(name)s" was expired',
            dict(
                id=logger.id,
                celery_task_id=logger.celery_task_id or '',
                name=logger.name,
            ),
            extra=dict(
                id=logger.id,
                celery_task_invocation_task_id=logger.celery_task_id or '',
                celery_task_invocation_name=logger.name,
                celery_task_invocation_queue=logger.queue_name,
                celery_task_invocation_start=logger.start,
                celery_task_invocation_stop=logger.stop,
                celery_task_invocation_time=logger.time,
                celery_task_invocation_state='EXPIRED',
                celery_task_invocation_release=logger.release
            ),
        )

    def celery_task_run_started(self, logger):
        celery_logger.info(
            'Celery task "%(id)s" with celery id "%(celery_task_id)s" and name "%(name)s" was started',
            dict(
                id=logger.id,
                celery_task_id=logger.celery_task_id,
                name=logger.name,
            ),
            extra=dict(
                id=logger.id,
                celery_task_name=logger.name,
                celery_task_queue=logger.queue_name,
                celery_task_start=logger.start,
                celery_task_invocation_state='ACTIVE',
                celery_task_attempt=logger.retries,
                celery_task_release=logger.release,
                celery_task_waiting_time=logger.waiting_time
            ),
        )

    def celery_task_run_succeeded(self, logger):
        celery_logger.info(
            'Celery task "%(id)s" with celery id "%(celery_task_id)s" and name "%(name)s" was successful',
            dict(
                id=logger.id,
                celery_task_id=logger.celery_task_id,
                name=logger.name,
            ),
            extra=dict(
                id=logger.id,
                celery_task_name=logger.name,
                celery_task_queue=logger.queue_name,
                celery_task_start=logger.start,
                celery_task_invocation_state='SUCCEEDED',
                celery_task_attempt=logger.retries,
                celery_task_stop=logger.stop,
                celery_task_time=logger.time,
                celery_task_release=logger.release,
                celery_task_waiting_time=logger.waiting_time
            )
        )

    def celery_task_run_failed(self, logger):
        celery_logger.error(
            'Celery task "%(id)s" with celery id "%(celery_task_id)s" and name "%(name)s" failed',
            dict(
                id=logger.id,
                celery_task_id=logger.celery_task_id,
                name=logger.name,
            ),
            extra=dict(
                id=logger.id,
                celery_task_name=logger.name,
                celery_task_queue=logger.queue_name,
                celery_task_start=logger.start,
                celery_task_invocation_state='FAILED',
                celery_task_attempt=logger.retries,
                celery_task_stop=logger.stop,
                celery_task_time=logger.time,
                celery_task_release=logger.release,
                celery_task_waiting_time=logger.waiting_time
            )
        )

    def celery_task_run_retried(self, logger):
        celery_logger.warning(
            'Celery task "%(id)s" with celery id "%(celery_task_id)s" and name "%(name)s" was repeated',
            dict(
                id=logger.id,
                celery_task_id=logger.celery_task_id,
                name=logger.name,
            ),
            extra=dict(
                id=logger.id,
                celery_task_name=logger.name,
                celery_task_queue=logger.queue_name,
                celery_task_start=logger.start,
                celery_task_invocation_state='RETRIED',
                celery_task_attempt=logger.retries,
                celery_task_stop=logger.stop,
                celery_task_time=logger.time,
                celery_task_release=logger.release,
                celery_task_waiting_time=logger.waiting_time
            )
        )

    def celery_task_run_output_updated(self, logger):
        """Output is not stored into log"""



# Generated by Django 3.1 on 2021-11-29 10:42

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('security_backends_sql', '0002_auto_20211119_1940'),
    ]

    operations = [
        migrations.AddField(
            model_name='celerytaskinvocationlog',
            name='parent_log',
            field=models.CharField(blank=True, db_index=True, max_length=250, null=True),
        ),
        migrations.AddField(
            model_name='celerytaskrunlog',
            name='parent_log',
            field=models.CharField(blank=True, db_index=True, max_length=250, null=True),
        ),
        migrations.AddField(
            model_name='commandlog',
            name='parent_log',
            field=models.CharField(blank=True, db_index=True, max_length=250, null=True),
        ),
        migrations.AddField(
            model_name='inputrequestlog',
            name='parent_log',
            field=models.CharField(blank=True, db_index=True, max_length=250, null=True),
        ),
        migrations.AddField(
            model_name='outputrequestlog',
            name='parent_log',
            field=models.CharField(blank=True, db_index=True, max_length=250, null=True),
        ),
    ]



from django.utils.translation import gettext_lazy as _

from enum import Enum

from enumfields import Choice, IntegerChoicesEnum


class RequestLogState(IntegerChoicesEnum):

    INCOMPLETE = Choice(0, _('Incomplete'))
    DEBUG = Choice(10, _('Debug'))
    INFO = Choice(20, _('Info'))
    WARNING = Choice(30, _('Warning'))
    ERROR = Choice(40, _('Error'))


class CommandState(IntegerChoicesEnum):

    ACTIVE = Choice(1, _('Active'))
    SUCCEEDED = Choice(2, _('Succeeded'))
    FAILED = Choice(3, _('Failed'))


class CeleryTaskInvocationLogState(IntegerChoicesEnum):

    WAITING = Choice(1, _('Waiting'))
    TRIGGERED = Choice(7, _('Triggered'))
    ACTIVE = Choice(2, _('Active'))
    SUCCEEDED = Choice(3, _('Succeeded'))
    FAILED = Choice(4, _('Failed'))
    EXPIRED = Choice(6, _('Expired'))
    TIMEOUT = Choice(8, _('Timeout'))
    IGNORED = Choice(9, _('Ignored'))
    DUPLICATE = Choice(10, _('Duplicate'))


class CeleryTaskRunLogState(IntegerChoicesEnum):

    ACTIVE = Choice(1, _('Active'))
    SUCCEEDED = Choice(2, _('Succeeded'))
    FAILED = Choice(3, _('Failed'))
    RETRIED = Choice(4, _('Retried'))
    EXPIRED = Choice(5, _('Expired'))


class InputRequestSlug(str, Enum):

    UNSUCCESSFUL_LOGIN_REQUEST = 'UNSUCCESSFUL_LOGIN_REQUEST'
    SUCCESSFUL_LOGIN_REQUEST = 'SUCCESSFUL_LOGIN_REQUEST'
    UNSUCCESSFUL_2FA_CODE_VERIFICATION_REQUEST = 'UNSUCCESSFUL_2FA_CODE_VERIFICATION_REQUEST'
    SUCCESSFUL_2FA_CODE_VERIFICATION_REQUEST = 'SUCCESSFUL_2FA_CODE_VERIFICATION_REQUEST'


class LoggerName(str, Enum):

    INPUT_REQUEST = 'input-request',
    OUTPUT_REQUEST = 'output-request',
    COMMAND = 'command',
    CELERY_TASK_INVOCATION = 'celery-task-invocation'
    CELERY_TASK_RUN = 'celery-task-run'



from elasticsearch_dsl import Q

from security.backends.reader import BaseBackendReader

from .models import InputRequestLog, get_log_model_from_logger_name, get_key_from_object


class ElasticsearchBackendReader(BaseBackendReader):

    def get_count_input_requests(self, from_time, ip=None, path=None, view_slug=None, slug=None, method=None,
                                 exclude_log_id=None):
        q = Q('range', start={'gte': from_time})

        if ip is not None:
            q &= Q('term', ip=ip)
        if path is not None:
            q &= Q('term', path=path)
        if method is not None:
            q &= Q('term', method=method)
        if view_slug is not None:
            q &= Q('term', view_slug=view_slug)
        if slug is not None:
            q &= Q('term', slug=slug)
        if exclude_log_id is not None:
            q &= ~Q('ids', **{'values': [exclude_log_id]})
        return InputRequestLog.search().filter(q).count()

    def get_logs_related_with_object(self, logger_name, related_object):
        related_object_key = get_key_from_object(related_object)
        return list(get_log_model_from_logger_name(logger_name).search().filter(
            Q('term', related_objects=related_object_key)
        ))



from uuid import uuid4

from django.test.utils import override_settings

from .models import CommandLog, CeleryTaskRunLog, CeleryTaskInvocationLog, InputRequestLog, OutputRequestLog


class store_elasticsearch_log(override_settings):

    def __init__(self, **kwargs):
        super().__init__(
            SECURITY_BACKEND_WRITERS=('elasticsearch',), SECURITY_ELASTICSEARCH_AUTO_REFRESH=True, **kwargs
        )

    def enable(self):
        from .connection import set_connection

        super().enable()
        uuid = uuid4()
        set_connection()
        for document_class in (CommandLog, CeleryTaskRunLog, CeleryTaskInvocationLog,
                               InputRequestLog, OutputRequestLog):
            document_class._index._name = f'{uuid}.{document_class._index._name}'
            document_class.init()

    def disable(self):
        for document_class in (CommandLog, CeleryTaskRunLog, CeleryTaskInvocationLog,
                               InputRequestLog, OutputRequestLog):
            document_class._index.delete()
            document_class._index._name = document_class._index._name.split('.', 1)[1]
        super().disable()



from django.contrib.auth import get_user_model
from django.db import router
from django.db.models import TextField
from django.db.models.functions import Cast
from django.utils.translation import ugettext
from django.contrib.contenttypes.models import ContentType

from elasticsearch_dsl import Q

from pyston.contrib.elasticsearch.filters import ElasticsearchFilter
from pyston.filters.utils import OperatorSlug
from pyston.filters.exceptions import FilterValueError

from is_core.contrib.elasticsearch.filters import CoreElasticsearchFilterManagerFilterManager

from security.config import settings
from security.backends.elasticsearch.models import get_key_from_content_type_object_id_and_model_db


class UsernameUserFilter(ElasticsearchFilter):

    allowed_operators = (OperatorSlug.CONTAINS,)

    def clean_value(self, value, operator_slug, request):
        user_model = get_user_model()
        queryset = user_model.objects.filter(
            **{'{}__contains'.format(user_model.USERNAME_FIELD): value}
        ).annotate(
            str_id=Cast('id', output_field=TextField())
        )

        if queryset[:settings.ELASTICSEARCH_MAX_NUMBER_OF_TERMS].count() > settings.ELASTICSEARCH_MAX_NUMBER_OF_TERMS:
            raise FilterValueError(ugettext('Too many users found for specified username.'))

        return list(queryset.values_list('str_id', flat=True))

    def get_filter_term(self, value, operator_slug, request):
        return Q('terms', **{'user_id': value})


class EnumElasticsearchFilter(ElasticsearchFilter):

    allowed_operators = (OperatorSlug.EQ,)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._enum = self.field._enum
        self.choices = [(None, '')] + [
            (choice.name, choice.label) for choice in self._enum
        ]

    def clean_value(self, value, operator_slug, request):
        try:
            return self._enum[value]
        except KeyError:
            raise FilterValueError(
                ugettext('Invalid value. Please use one of the following values: {}.').format(
                    ', '.join([a.name for a in self._enum])
                )
            )

    def get_filter_term(self, value, operator_slug, request):
        return Q('term', **{self.get_full_filter_key(): value.name})


class SecurityElasticsearchFilterManager(CoreElasticsearchFilterManagerFilterManager):

    filter_by_field_name = {
        'enum': EnumElasticsearchFilter,
        **CoreElasticsearchFilterManagerFilterManager.filter_by_field_name,
    }


class RelatedObjectsFilter(ElasticsearchFilter):

    allowed_operators = (OperatorSlug.IN,)

    def clean_value(self, value, operator_slug, request):
        cleaned_values = []
        for v in value:
            try:
                content_type_id, object_id = v.split('|')

                content_type = ContentType.objects.get(pk=content_type_id)
                model_db = router.db_for_write(content_type.model_class())

                cleaned_values.append(
                    get_key_from_content_type_object_id_and_model_db(model_db, content_type_id, object_id)
                )
            except (ValueError, ContentType.DoesNotExist):
                raise FilterValueError(ugettext('Invalid value.'))
        return cleaned_values

    def get_filter_term(self, value, operator_slug, request):
        return Q('terms', related_objects=value)



from django.utils.timezone import now

from security.enums import LoggerName
from security.logging.common import SecurityLogger
from security.backends.signals import (
    celery_task_invocation_started, celery_task_invocation_triggered, celery_task_invocation_ignored,
    celery_task_invocation_timeout, celery_task_invocation_expired, celery_task_run_started, celery_task_run_succeeded,
    celery_task_run_failed, celery_task_run_retried, celery_task_run_output_updated, celery_task_invocation_succeeded,
    celery_task_invocation_failed, celery_task_invocation_duplicate
)


class CeleryInvocationLogger(SecurityLogger):

    logger_name = LoggerName.CELERY_TASK_INVOCATION
    store = False

    def __init__(self, name=None, queue_name=None, input=None, task_args=None, task_kwargs=None, applied_at=None,
                 is_async=None, is_unique=None, is_on_commit=None, triggered_at=None, stale_at=None,
                 estimated_time_of_first_arrival=None, expires_at=None, celery_task_id=None, **kwargs):
        super().__init__(**kwargs)
        self.name = name
        self.input = input
        self.queue_name = queue_name
        self.task_args = task_args
        self.task_kwargs = task_kwargs
        self.applied_at = applied_at
        self.is_async = is_async
        self.is_unique = is_unique
        self.is_on_commit = is_on_commit
        self.triggered_at = triggered_at
        self.stale_at = stale_at
        self.estimated_time_of_first_arrival = estimated_time_of_first_arrival
        self.expires_at = expires_at
        self.celery_task_id = celery_task_id

    def log_started(self, name, queue_name, task_args, task_kwargs, applied_at,
                    is_async, is_unique, is_on_commit):
        task_input = []
        if task_args:
            task_input += [str(v) for v in task_args]
        if task_kwargs:
            task_input += ['{}={}'.format(k, v) for k, v in task_kwargs.items()]
        self.name = name
        self.queue_name = queue_name
        self.input = ', '.join(task_input)
        self.task_args = list(task_args)
        self.task_kwargs = task_kwargs
        self.applied_at = applied_at
        self.is_async = is_async
        self.is_unique = is_unique
        self.is_on_commit = is_on_commit
        self.start = now()
        celery_task_invocation_started.send(
            sender=CeleryInvocationLogger,
            logger=self,
        )

    def log_triggered(self, triggered_at, stale_at, estimated_time_of_first_arrival,
                      expires_at, celery_task_id):
        self.triggered_at = triggered_at
        self.stale_at = stale_at
        self.estimated_time_of_first_arrival = estimated_time_of_first_arrival
        self.expires_at = expires_at
        self.celery_task_id = celery_task_id
        celery_task_invocation_triggered.send(
            sender=CeleryInvocationLogger,
            logger=self,
        )

    def log_unique(self, triggered_at, stale_at, estimated_time_of_first_arrival,
                   expires_at, celery_task_id):
        self.triggered_at = triggered_at
        self.stale_at = stale_at
        self.estimated_time_of_first_arrival = estimated_time_of_first_arrival
        self.expires_at = expires_at
        self.celery_task_id = celery_task_id
        celery_task_invocation_duplicate.send(
            sender=CeleryInvocationLogger,
            logger=self,
        )

    def log_ignored(self):
        self.stop = now()
        celery_task_invocation_ignored.send(
            sender=CeleryInvocationLogger,
            logger=self,
        )

    def log_timeout(self):
        self.stop = now()
        celery_task_invocation_timeout.send(
            sender=CeleryInvocationLogger,
            logger=self,
        )

    def log_expired(self):
        self.stop = now()
        celery_task_invocation_expired.send(
            sender=CeleryInvocationLogger,
            logger=self,
        )

    def log_succeeded(self):
        self.stop = now()
        celery_task_invocation_succeeded.send(
            sender=CeleryInvocationLogger,
            logger=self,
        )

    def log_failed(self):
        self.stop = now()
        celery_task_invocation_failed.send(
            sender=CeleryInvocationLogger,
            logger=self,
        )


class CeleryTaskRunLogger(SecurityLogger):

    logger_name = LoggerName.CELERY_TASK_RUN

    def __init__(self, name=None, queue_name=None, input=None, task_args=None, task_kwargs=None, retries=None,
                 waiting_time=None, celery_task_id=None, result=None, output=None, estimated_time_of_next_retry=None,
                 **kwargs):
        super().__init__(**kwargs)
        self.name = name
        self.input = input
        self.queue_name = queue_name
        self.task_args = task_args
        self.task_kwargs = task_kwargs
        self.retries = retries
        self.waiting_time = waiting_time
        self.celery_task_id = celery_task_id
        self.result = result
        self.output = output
        self.estimated_time_of_next_retry = estimated_time_of_next_retry

    def log_started(self, name, queue_name, task_args, task_kwargs, celery_task_id, retries, trigger_time):
        task_input = []
        if task_args:
            task_input += [str(v) for v in task_args]
        if task_kwargs:
            task_input += ['{}={}'.format(k, v) for k, v in task_kwargs.items()]
        self.start = now()
        self.name = name
        self.celery_task_id = celery_task_id
        self.queue_name = queue_name
        self.input = ', '.join(task_input)
        self.task_args = list(task_args)
        self.task_kwargs = task_kwargs
        self.retries = retries
        self.waiting_time = (self.start - trigger_time).total_seconds()
        celery_task_run_started.send(
            sender=CeleryTaskRunLogger,
            logger=self,
        )

    def log_succeeded(self, result):
        self.result = result
        self.stop = now()
        celery_task_run_succeeded.send(
            sender=CeleryTaskRunLogger,
            logger=self,
        )

    def log_failed(self, ex_tb):
        self.error_message = ex_tb
        self.stop = now()
        celery_task_run_failed.send(
            sender=CeleryTaskRunLogger,
            logger=self,
        )

    def log_retried(self, ex_tb, estimated_time_of_next_retry):
        self.error_message = ex_tb
        self.stop = now()
        self.estimated_time_of_next_retry = estimated_time_of_next_retry
        celery_task_run_retried.send(
            sender=CeleryTaskRunLogger,
            logger=self,
        )

    def log_output_updated(self, output):
        self.output = output
        celery_task_run_output_updated.send(sender=CeleryTaskRunLogger, logger=self)



import copy

from threading import local

from uuid import uuid4

from contextlib import ContextDecorator

from django.conf import settings as django_settings

from security.config import settings
from security.utils import get_object_triple


undefined = object()


class SecurityLogger(ContextDecorator, local):

    loggers = []
    logger_name = None
    store = True

    def __init__(self, id=None, parent_log=undefined, related_objects=None, slug=None, extra_data=None,
                 start=None, stop=None, error_message=None, time=None, release=None):
        self.id = id or (uuid4() if self.logger_name else None)
        self.parent = SecurityLogger.loggers[-1] if SecurityLogger.loggers else None

        self.related_objects = set()
        if related_objects:
            self.add_related_objects(*related_objects)

        self.start = start
        self.stop = stop
        self.error_message = error_message
        self.release = release or settings.RELEASE

        self.slug = slug
        if self.parent:
            self.related_objects |= self.parent.related_objects
            if not self.slug:
                self.slug = self.parent.slug
        parent_with_id = self._get_parent_with_id()
        self.parent_log = (
            '{}|{}'.format(parent_with_id.logger_name.value, parent_with_id.id) if parent_with_id else None
        ) if parent_log is undefined else parent_log

        self._extra_data = extra_data
        if self._extra_data is None:
            self._extra_data = self.parent.extra_data if self.parent else {}

        if self.store:
            SecurityLogger.loggers.append(self)

            if 'reversion' in django_settings.INSTALLED_APPS:
                from reversion.signals import post_revision_commit

                post_revision_commit.connect(self._post_revision_commit, weak=False)
        self.backend_logs = {}
        self.stream = None

    def _get_parent_with_id(self):
        parent = self.parent
        while parent and not parent.id:
            parent = parent.parent
        return parent

    @property
    def time(self):
        return (self.stop - self.start).total_seconds() if self.start and self.stop else None

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.store:
            self.close()

    def set_slug(self, slug):
        self.slug = slug

    def add_related_objects(self, *related_objects):
        self.related_objects |= set(get_object_triple(obj) for obj in related_objects)

    @property
    def extra_data(self):
        return copy.deepcopy(self._extra_data)

    def update_extra_data(self, data):
        self._extra_data.update(data)

    def close(self):
        if not SecurityLogger.loggers or SecurityLogger.loggers[-1] != self:
            raise RuntimeError('Log already finished')

        SecurityLogger.loggers.pop()
        if 'reversion' in django_settings.INSTALLED_APPS:
            from reversion.signals import post_revision_commit

            post_revision_commit.disconnect(self._post_revision_commit)

    def _post_revision_commit(self, **kwargs):
        """
        Called as a post save of revision model of the reversion library.
        If log context manager is active input logged request, command
        log or celery task run log is joined with revision via related objects.
        """
        reversion_data = self._extra_data['reversion'] = self._extra_data.get('reversion', {
            'revisions': [],
            'total_count': 0
        })
        if reversion_data['total_count'] < settings.LOG_MAX_REVISIONS_COUNT:
            reversion_data['revisions'].append({
                'id': kwargs['revision_id'] if 'revision_id' in kwargs else kwargs['revision'].id
            })
        reversion_data['total_count'] += 1

    def to_dict(self):
        return dict(
            extra_data=self.extra_data,
            time=self.time,
            **{k: v for k, v in self.__dict__.items() if k not in {
                'backend_logs', 'stream', 'store', 'logger_name', 'loggers', 'parent'
            } and not k.startswith('_')}
        )


def get_last_logger(name):
    for logger in SecurityLogger.loggers[::-1]:
        if logger.logger_name == name:
            return logger
    return None



import logging

from datetime import timedelta

from django.core.management.base import OutputWrapper
from django.db import transaction
from django_celery_extensions.task import DjangoTask, ResultWrapper

from celery.utils.time import maybe_iso8601

from chamber.utils.transaction import pre_commit, in_atomic_block

from security.config import settings
from security.utils import LogStringIO, deserialize_data, serialize_data
from security.logging.celery.logger import CeleryTaskRunLogger, CeleryInvocationLogger


logger = logging.getLogger(__name__)


class LoggedResultWrapper(ResultWrapper):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.invocation_logger = CeleryInvocationLogger(
            self._invocation_id, related_objects=self._options.pop('related_objects', [])
        )

    def on_apply(self):
        self.invocation_logger.log_started(
            name=self._task.name,
            queue_name=self._options['queue'],
            task_args=self._args or [],
            task_kwargs=self._kwargs or {},
            applied_at=self._options['apply_time'],
            is_async=self._options['is_async'],
            is_unique=self._task.unique,
            is_on_commit=self._options['is_on_commit']
        )

    def on_trigger(self):
        stale_time_limit = self._options.get('stale_time_limit')
        self.invocation_logger.log_triggered(
            triggered_at=self._options.get('trigger_time'),
            stale_at=(
                self._options.get('trigger_time') + timedelta(seconds=stale_time_limit)
                if stale_time_limit is not None else None
            ),
            estimated_time_of_first_arrival=self._options.get('eta'),
            expires_at=self._options.get('expires'),
            celery_task_id=self._options.get('task_id')
        )

    def on_unique(self):
        stale_time_limit = self._options.get('stale_time_limit')
        self.invocation_logger.log_unique(
            triggered_at=self._options.get('trigger_time'),
            stale_at=(
                self._options.get('trigger_time') + timedelta(seconds=stale_time_limit)
                if stale_time_limit is not None else None
            ),
            estimated_time_of_first_arrival=self._options.get('eta'),
            expires_at=self._options.get('expires'),
            celery_task_id=self._options.get('task_id')
        )

    def on_ignored(self):
        self.invocation_logger.log_ignored()

    def on_timeout(self):
        self.invocation_logger.log_timeout()


class LoggedTask(DjangoTask):

    abstract = True
    result_wrapper_class = LoggedResultWrapper

    def on_invocation_apply(self, invocation_id, args, kwargs, options, result):
        _super = super()

        def _on_invocation_apply():
            _super.on_invocation_apply(invocation_id, args, kwargs, options, result)

        if options.get('is_on_commit') and in_atomic_block():
            if settings.TASK_USE_PRE_COMMIT:
                pre_commit(_on_invocation_apply, using=options.get('using'))
            else:
                transaction.on_commit(_on_invocation_apply, using=options.get('using'))
        else:
            _on_invocation_apply()

    def on_task_start(self, task_id, args, kwargs):
        super().on_task_start(task_id, args, kwargs)
        run_logger = CeleryTaskRunLogger()
        self.request.run_logger = run_logger
        self.request.invocation_logger = CeleryInvocationLogger(
            **deserialize_data(self._get_header_from_request('invocation_logger_data'))
        )
        run_logger.log_started(
            name=self.name,
            queue_name=self.request.delivery_info.get('routing_key'),
            task_args=args,
            task_kwargs=kwargs,
            celery_task_id=task_id,
            retries=self.request.retries,
            trigger_time=maybe_iso8601(self._get_header_from_request('trigger_time'))
        )
        self.request.output_stream = LogStringIO(
            flush_callback=lambda output_stream: run_logger.log_output_updated(output_stream.getvalue())
        )

    def on_task_retry(self, task_id, args, kwargs, exc, eta):
        super().on_task_retry(task_id, args, kwargs, exc, eta)
        try:
            self.request.output_stream.close()
            self.request.run_logger.log_retried(
                ex_tb=str(exc),
                estimated_time_of_next_retry=eta
            )
        finally:
            self.request.run_logger.close()

    def on_task_success(self, task_id, args, kwargs, retval):
        super().on_task_success(task_id, args, kwargs, retval)
        try:
            self.request.output_stream.close()
            self.request.run_logger.log_succeeded(result=retval)
            self.request.invocation_logger.log_succeeded()
        finally:
            self.request.run_logger.close()

    def on_task_failure(self, task_id, args, kwargs, exc, einfo):
        super().on_task_failure(task_id, args, kwargs, exc, einfo)
        log_id = None
        if hasattr(self.request, 'run_logger'):
            log_id = self.request.run_logger.id
            try:
                self.request.output_stream.close()
                self.request.run_logger.log_failed(ex_tb=str(exc))
                if not self.acks_late or self.acks_on_failure_or_timeout:
                    # Task will be started again invocation can't be logged
                    self.request.invocation_logger.log_failed()
            finally:
                self.request.run_logger.close()

        logger.error(
            f'Task with name {self.name} raised exception',
            extra={
                'einfo': str(einfo),
                'exception': str(exc),
                'task_id': task_id,
                'log_id': log_id
            }
        )

    def expire_invocation(self, invocation_log_id, args, kwargs, logger_data):
        with CeleryInvocationLogger(id=invocation_log_id, **logger_data) as invocation_logger:
            invocation_logger.log_expired()

        logger.error(
            f'Task with name {self.name} was expired',
            extra={
                'invocation_log_id': invocation_log_id
            }
        )

    def _apply_and_get_result(self, result, args, kwargs, invocation_id, is_async=False, headers=None, **options):
        headers = {} if headers is None else headers
        headers['invocation_logger_data'] = serialize_data(result.invocation_logger.to_dict())
        return super()._apply_and_get_result(
            result, args, kwargs, invocation_id, is_async=is_async, headers=headers, **options
        )

    def retry(self, headers=None, **options):
        headers = {} if headers is None else headers
        headers['invocation_logger_data'] = serialize_data(self.request.invocation_logger.to_dict())
        return super().retry(headers=headers, **options)

    @property
    def run_logger(self):
        return None if not self.request else self.request.run_logger

    @property
    def stdout(self):
        return OutputWrapper(self.request.output_stream)

    @property
    def stderr(self):
        return OutputWrapper(self.request.output_stream)

    @property
    def task_log(self):
        return self.request.celery_task_log

    def get_command_kwargs(self):
        return dict(
            stdout=self.request.output_stream,
            stderr=self.request.output_stream,
        )



from uuid import uuid4
from unittest import mock
import responses
from datetime import timedelta

from django import get_version
from django.contrib.contenttypes.models import ContentType
from django.contrib.auth.models import User
from django.db import transaction
from django.test import override_settings
from django.utils.timezone import now

from freezegun import freeze_time

from celery.exceptions import TimeoutError

from germanium.decorators import data_consumer
from germanium.test_cases.client import ClientTestCase
from germanium.tools import (
    assert_equal, assert_false, assert_http_not_found, assert_http_ok, assert_http_redirect,
    assert_http_too_many_requests, assert_in, assert_is_none, assert_is_not_none, assert_not_in,
    assert_raises, assert_true, assert_equal_model_fields, capture_commit_callbacks, assert_length_equal,
    all_eq_obj, not_none_eq_obj
)

from chamber.shortcuts import change_and_save

from django_celery_extensions.task import get_django_command_task

from security.config import settings
from security.decorators import log_with_data
from security.enums import CeleryTaskInvocationLogState, CeleryTaskRunLogState, LoggerName
from security.backends.sql.models import CeleryTaskRunLog as SQLCeleryTaskRunLog
from security.backends.sql.models import CeleryTaskInvocationLog as SQLCeleryTaskInvocationLog
from security.backends.elasticsearch.tests import store_elasticsearch_log
from security.backends.elasticsearch.models import CeleryTaskRunLog as ElasticsearchCeleryTaskRunLog
from security.backends.elasticsearch.models import CeleryTaskInvocationLog as ElasticsearchCeleryTaskInvocationLog
from security.backends.signals import (
    celery_task_invocation_started, celery_task_invocation_expired, celery_task_invocation_triggered,
    celery_task_invocation_ignored, celery_task_invocation_timeout, celery_task_run_started, celery_task_run_failed,
    celery_task_run_retried, celery_task_run_succeeded, celery_task_run_output_updated
)
from security.backends.testing import capture_security_logs
from security.backends.reader import get_logs_related_with_object
from security.utils import get_object_triple

from apps.test_security.tasks import (
    error_task, retry_task, sum_task, unique_task, ignored_after_success_task,
    acks_late_with_no_failure_or_timeout_acknowledgement
)

from .base import BaseTestCaseMixin, TRUNCATION_CHAR, test_call_command, assert_equal_logstash, assert_equal_log_data


@override_settings(SECURITY_BACKEND_WRITERS={})
class CeleryLogTestCase(BaseTestCaseMixin, ClientTestCase):

    @data_consumer('create_user')
    @override_settings(SECURITY_BACKEND_READER='testing')
    def test_sum_celery_task_should_be_logged(self, user):
        expected_invocation_started_data = {
            'name': 'sum_task',
            'queue_name': 'default',
            'input': '5, 8',
            'task_args': [5, 8],
            'task_kwargs': {},
            'applied_at': not_none_eq_obj,
            'is_async': False,
            'is_unique': False,
            'is_on_commit': False,
            'start': not_none_eq_obj,
        }
        expected_invocation_triggered_data = {
            **expected_invocation_started_data,
            'triggered_at': not_none_eq_obj,
            'stale_at': None,
            'estimated_time_of_first_arrival': not_none_eq_obj,
            'expires_at': None,
            'celery_task_id': not_none_eq_obj
        }

        expected_run_started_data = {
            'name': 'sum_task',
            'queue_name': None,
            'input': '5, 8',
            'task_args': [5, 8],
            'task_kwargs': {},
            'start': not_none_eq_obj,
            'retries': 0,
            'celery_task_id': not_none_eq_obj,
            'waiting_time': not_none_eq_obj,
        }
        expected_run_succeeded_data = {
            **expected_run_started_data,
            'stop': not_none_eq_obj,
            'result': 13,
        }

        with capture_security_logs() as logged_data:
            sum_task.apply(args=(5, 8), related_objects=[user])
            assert_length_equal(logged_data.celery_task_invocation_started, 1)
            assert_length_equal(logged_data.celery_task_invocation_triggered, 1)
            assert_length_equal(logged_data.celery_task_run_started, 1)
            assert_length_equal(logged_data.celery_task_run_succeeded, 1)
            assert_length_equal(logged_data.celery_task_run_failed, 0)
            assert_length_equal(logged_data.celery_task_run_retried, 0)
            assert_equal_log_data(logged_data.celery_task_invocation_started[0], expected_invocation_started_data)
            assert_equal_log_data(logged_data.celery_task_invocation_triggered[0], expected_invocation_triggered_data)
            assert_equal_log_data(logged_data.celery_task_run_started[0], expected_run_started_data)
            assert_equal_log_data(logged_data.celery_task_run_succeeded[0], expected_run_succeeded_data)
            assert_equal(logged_data.celery_task_invocation[0].related_objects, {get_object_triple(user)})
            assert_equal(
                get_logs_related_with_object(LoggerName.CELERY_TASK_INVOCATION, user),
                [logged_data.celery_task_invocation[0]]
            )

    def test_error_celery_task_should_be_logged(self):
        expected_invocation_started_data = {
            'name': 'error_task',
            'queue_name': 'default',
            'input': '',
            'task_args': [],
            'task_kwargs': {},
            'applied_at': not_none_eq_obj,
            'is_async': False,
            'is_unique': False,
            'is_on_commit': False,
            'start': not_none_eq_obj,
        }
        expected_invocation_triggered_data = {
            **expected_invocation_started_data,
            'triggered_at': not_none_eq_obj,
            'stale_at': not_none_eq_obj,
            'estimated_time_of_first_arrival': not_none_eq_obj,
            'expires_at': None,
            'celery_task_id': not_none_eq_obj,
        }

        expected_run_started_data = {
            'name': 'error_task',
            'queue_name': None,
            'input': '',
            'task_args': [],
            'task_kwargs': {},
            'start': not_none_eq_obj,
            'retries': 0,
            'celery_task_id': not_none_eq_obj,
            'waiting_time': not_none_eq_obj,
        }
        expected_run_failed_data = {
            **expected_run_started_data,
            'stop': not_none_eq_obj,
            'error_message': not_none_eq_obj,
        }
        with capture_security_logs() as logged_data:
            error_task.apply()
            assert_length_equal(logged_data.celery_task_invocation_started, 1)
            assert_length_equal(logged_data.celery_task_invocation_triggered, 1)
            assert_length_equal(logged_data.celery_task_invocation_failed, 1)
            assert_length_equal(logged_data.celery_task_run_started, 1)
            assert_length_equal(logged_data.celery_task_run_succeeded, 0)
            assert_length_equal(logged_data.celery_task_run_failed, 1)
            assert_length_equal(logged_data.celery_task_run_retried, 0)
            assert_equal_log_data(logged_data.celery_task_invocation_started[0], expected_invocation_started_data)
            assert_equal_log_data(logged_data.celery_task_invocation_triggered[0], expected_invocation_triggered_data)
            assert_equal_log_data(logged_data.celery_task_run_started[0], expected_run_started_data)
            assert_equal_log_data(logged_data.celery_task_run_failed[0], expected_run_failed_data)

    def test_retry_celery_task_should_be_logged(self):
        expected_invocation_started_data = {
            'name': 'retry_task',
            'queue_name': 'default',
            'input': '',
            'task_args': [],
            'task_kwargs': {},
            'applied_at': not_none_eq_obj,
            'is_async': False,
            'is_unique': False,
            'is_on_commit': False,
            'start': not_none_eq_obj,
        }
        expected_invocation_triggered_data = {
            **expected_invocation_started_data,
            'triggered_at': not_none_eq_obj,
            'stale_at': None,
            'estimated_time_of_first_arrival': not_none_eq_obj,
            'expires_at': None,
            'celery_task_id': not_none_eq_obj,
        }

        expected_run_started_data = {
            'name': 'retry_task',
            'queue_name': None,
            'input': '',
            'task_args': [],
            'task_kwargs': {},
            'start': not_none_eq_obj,
            'retries': not_none_eq_obj,
            'celery_task_id': not_none_eq_obj,
            'waiting_time': not_none_eq_obj,
        }
        expected_run_retried_data = {
            **expected_run_started_data,
            'stop': not_none_eq_obj,
            'error_message': not_none_eq_obj,
            'estimated_time_of_next_retry': not_none_eq_obj,
        }

        with capture_security_logs() as logged_data:
            retry_task.apply()

            assert_length_equal(logged_data.celery_task_invocation_started, 1)
            assert_length_equal(logged_data.celery_task_invocation_triggered, 1)
            assert_length_equal(logged_data.celery_task_run_started, 6)
            assert_length_equal(logged_data.celery_task_run_succeeded, 1)
            assert_length_equal(logged_data.celery_task_run_failed, 0)
            assert_length_equal(logged_data.celery_task_run_retried, 5)
            assert_equal_log_data(logged_data.celery_task_invocation_started[0], expected_invocation_started_data)
            assert_equal_log_data(logged_data.celery_task_invocation_triggered[0], expected_invocation_triggered_data)
            assert_equal_log_data(logged_data.celery_task_run_started[0], expected_run_started_data)
            assert_equal_log_data(logged_data.celery_task_run_retried[0], expected_run_retried_data)

    @override_settings(DJANGO_CELERY_EXTENSIONS_DEFAULT_TASK_STALE_TIME_LIMIT=30, SECURITY_TASK_USE_ON_SUCCESS=True)
    def test_ignored_after_success_celery_task_should_be_logged(self):
        with capture_security_logs() as logged_data:
            with capture_commit_callbacks(execute_on_commit=True):
                ignored_after_success_task.apply_async_on_commit()
                ignored_after_success_task.apply_async_on_commit()

            assert_length_equal(logged_data.celery_task_invocation_started, 2)
            assert_length_equal(logged_data.celery_task_invocation_triggered, 1)
            assert_length_equal(logged_data.celery_task_invocation_ignored, 1)
            assert_length_equal(logged_data.celery_task_run_started, 1)

    @override_settings(DJANGO_CELERY_EXTENSIONS_DEFAULT_TASK_STALE_TIME_LIMIT=5)
    def test_apply_async_and_get_result_should_return_time_error_for_zero_timeout(self):
        with capture_security_logs() as logged_data:
            with assert_raises(TimeoutError):
                unique_task.apply_async_and_get_result(timeout=0)
            assert_length_equal(logged_data.celery_task_invocation_timeout, 1)

    @override_settings(SECURITY_TASK_USE_PRE_COMMIT=True)
    def test_task_should_be_logged_with_on_commit_signal(self):
        with capture_security_logs() as logged_data:
            with capture_commit_callbacks() as captured_commit_callbacks:
                sum_task.apply_async_on_commit(args=(5, 8))
            assert_length_equal(logged_data.celery_task_invocation_started, 0)
            assert_length_equal(logged_data.celery_task_invocation_triggered, 0)
            captured_commit_callbacks.execute_pre_commit()
            assert_length_equal(logged_data.celery_task_invocation_started, 1)
            assert_length_equal(logged_data.celery_task_invocation_triggered, 0)
            captured_commit_callbacks.execute_on_commit()
            assert_length_equal(logged_data.celery_task_invocation_started, 1)
            assert_length_equal(logged_data.celery_task_invocation_triggered, 1)

    @override_settings(SECURITY_TASK_USE_ON_SUCCESS=True)
    def test_unique_task_should_be_logged_as_duplicate_and_run_is_not_started(self):
        with mock.patch.object(unique_task, '_get_unique_task_id') as apply_method:
            unique_task_id = uuid4()
            apply_method.return_value = unique_task_id
            with capture_security_logs() as logged_data:
                unique_task.apply_async()
                assert_length_equal(logged_data.celery_task_invocation_duplicate, 1)
                assert_length_equal(logged_data.celery_task_run_started, 0)
                assert_equal(logged_data.celery_task_invocation_duplicate[0].celery_task_id, unique_task_id)

    def test_data_change_should_be_connected_with_celery_task_run_log(self):
        with capture_security_logs() as logged_data:
            get_django_command_task('create_user').apply_async()
            assert_equal(
                logged_data.celery_task_run[0].extra_data,
                {'reversion': {'revisions': [{'id': not_none_eq_obj}], 'total_count': 1}}
            )

    @override_settings(SECURITY_BACKEND_WRITERS={'sql'})
    @data_consumer('create_user')
    def test_sum_celery_task_should_be_logged_in_sql_backend(self, user):
        with log_with_data(related_objects=[user]):
            sum_task.apply(args=(5, 8), related_objects=[user])
            assert_equal(SQLCeleryTaskInvocationLog.objects.count(), 1)
            assert_equal(SQLCeleryTaskRunLog.objects.count(), 1)
            sql_celery_task_invocation_log = SQLCeleryTaskInvocationLog.objects.get()
            sql_celery_task_run_log = SQLCeleryTaskRunLog.objects.get()

            assert_equal_model_fields(
                sql_celery_task_invocation_log,
                state=CeleryTaskInvocationLogState.SUCCEEDED,
                name='sum_task',
                time=(sql_celery_task_invocation_log.stop - sql_celery_task_invocation_log.start).total_seconds(),
                stale_at=None,
                expires_at=None,
                input='5, 8',
                task_args=[5, 8],
                task_kwargs={},
                queue_name='default'
            )
            assert_is_not_none(sql_celery_task_invocation_log.celery_task_id)
            assert_equal([rel_obj.object for rel_obj in sql_celery_task_invocation_log.related_objects.all()], [user])
            assert_equal_model_fields(
                sql_celery_task_run_log,
                state=CeleryTaskRunLogState.SUCCEEDED,
                name='sum_task',
                time=(sql_celery_task_run_log.stop - sql_celery_task_run_log.start).total_seconds(),
                input='5, 8',
                task_args=[5, 8],
                task_kwargs={},
                queue_name=None,
                retries=0
            )

    @override_settings(SECURITY_BACKEND_WRITERS={'elasticsearch'})
    @data_consumer('create_user')
    def test_sum_celery_task_should_be_logged_in_elasticsearch_backend(self, user):
        with log_with_data(related_objects=[user]):
            with capture_security_logs() as logged_data:
                sum_task.apply(args=(5, 8), related_objects=[user])

                elasticsearch_celery_task_invocation_log = ElasticsearchCeleryTaskInvocationLog.get(
                    id=logged_data.celery_task_invocation[0].id
                )
                elasticsearch_celery_task_run_log = ElasticsearchCeleryTaskRunLog.get(
                    id=logged_data.celery_task_run[0].id
                )

                assert_equal_model_fields(
                    elasticsearch_celery_task_invocation_log,
                    state=CeleryTaskInvocationLogState.SUCCEEDED,
                    name='sum_task',
                    time=(
                        elasticsearch_celery_task_invocation_log.stop - elasticsearch_celery_task_invocation_log.start
                    ).total_seconds(),
                    stale_at=None,
                    expires_at=None,
                    input='5, 8',
                    task_args='[5, 8]',
                    task_kwargs='{}',
                    queue_name='default'
                )
                assert_is_not_none(elasticsearch_celery_task_invocation_log.celery_task_id)
                assert_equal(
                    [rel_obj for rel_obj in elasticsearch_celery_task_invocation_log.related_objects],
                    ['default|3|{}'.format(user.id)]
                )
                assert_equal_model_fields(
                    elasticsearch_celery_task_run_log,
                    state=CeleryTaskRunLogState.SUCCEEDED,
                    name='sum_task',
                    time=(
                        elasticsearch_celery_task_run_log.stop - elasticsearch_celery_task_run_log.start
                    ).total_seconds(),
                    input='5, 8',
                    task_args='[5, 8]',
                    task_kwargs='{}',
                    queue_name=None,
                    retries=0
                )

    @store_elasticsearch_log(SECURITY_ELASTICSEARCH_LOGSTASH_WRITER=True)
    @data_consumer('create_user')
    def test_sum_celery_task_should_be_logged_in_elasticsearch_backend_through_logstash(self, user):
        with log_with_data(related_objects=[user]):
            with capture_security_logs() as logged_data:
                with self.assertLogs('security.logstash', level='INFO') as cm:
                    sum_task.apply(args=(5, 8), related_objects=[user])

                    celery_task_invocation_log = logged_data.celery_task_invocation[0]
                    celery_task_run_log = logged_data.celery_task_run[0]

                    assert_equal(len(cm.output), 5)
                    start_invocation, trigger_invocation, start_run, success_run, success_invocation = cm.output

                    start_invocation_expected_data = {
                        'slug': None,
                        'release': None,
                        'related_objects': ['|'.join(str(v) for v in get_object_triple(user))],
                        'extra_data': {},
                        'parent_log': None,
                        'name': 'sum_task',
                        'queue_name': 'default',
                        'input': '5, 8',
                        'task_args': '[5, 8]',
                        'task_kwargs': '{}',
                        'applied_at': not_none_eq_obj,
                        'is_async': False,
                        'is_unique': False,
                        'is_on_commit': False,
                        'start': not_none_eq_obj,
                        'state': 'WAITING'
                    }
                    trigger_invocation_expected_data = {
                        **start_invocation_expected_data,
                        'triggered_at': not_none_eq_obj,
                        'state': 'TRIGGERED',
                        'stale_at': None,
                        'celery_task_id': not_none_eq_obj,
                        'estimated_time_of_first_arrival': not_none_eq_obj,
                        'expires_at': None,
                    }
                    success_invocation_expected_data ={
                        **trigger_invocation_expected_data,
                        'state': 'SUCCEEDED',
                        'stop': not_none_eq_obj,
                    }

                    start_run_expected_data = {
                        'slug': None,
                        'release': None,
                        'related_objects': ['|'.join(str(v) for v in get_object_triple(user))],
                        'extra_data': {},
                        'parent_log': None,
                        'name': 'sum_task',
                        'celery_task_id': not_none_eq_obj,
                        'queue_name': None,
                        'input': '5, 8',
                        'task_args': '[5, 8]',
                        'task_kwargs': '{}',
                        'start': not_none_eq_obj,
                        'retries': 0,
                        'waiting_time': not_none_eq_obj,
                        'state': 'ACTIVE',
                    }
                    success_run_expected_data = {
                        **start_run_expected_data,
                        'time': not_none_eq_obj,
                        'stop': not_none_eq_obj,
                        'state': 'SUCCEEDED',
                        'result': 13,
                    }

                    assert_equal_logstash(
                        start_invocation,
                        'security-celery-task-invocation-log',
                        0,
                        celery_task_invocation_log.id,
                        start_invocation_expected_data
                    )
                    assert_equal_logstash(
                        trigger_invocation,
                        'security-celery-task-invocation-log',
                        1,
                        celery_task_invocation_log.id,
                        trigger_invocation_expected_data
                    )

                    assert_equal_logstash(
                        start_run,
                        'security-celery-task-run-log',
                        0,
                        celery_task_run_log.id,
                        start_run_expected_data
                    )
                    assert_equal_logstash(
                        success_run,
                        'security-celery-task-run-log',
                        9999,
                        celery_task_run_log.id,
                        success_run_expected_data
                    )
                    assert_equal_logstash(
                        success_invocation,
                        'security-celery-task-invocation-log',
                        9999,
                        celery_task_invocation_log.id,
                        success_invocation_expected_data
                    )

    @override_settings(SECURITY_BACKEND_WRITERS={'logging'})
    def test_sum_celery_task_should_be_logged_in_logging_backend(self):
        with capture_security_logs() as logged_data:
            with self.assertLogs('security.celery', level='INFO') as cm:
                sum_task.apply(args=(5, 8))

                assert_equal(
                    cm.output,
                    [
                        f'INFO:security.celery:'
                        f'Celery task invocation "{logged_data.celery_task_invocation[0].id}" '
                        f'with celery id "{logged_data.celery_task_invocation[0].celery_task_id}" '
                        f'and name "sum_task" was invoked',
                        f'INFO:security.celery:Celery task "{logged_data.celery_task_run[0].id}" with '
                        f'celery id "{logged_data.celery_task_run[0].celery_task_id}" and name "sum_task" '
                        f'was started',
                        f'INFO:security.celery:Celery task "{logged_data.celery_task_run[0].id}" with '
                        f'celery id "{logged_data.celery_task_run[0].celery_task_id}" and name "sum_task" '
                        f'was successful',
                        f'INFO:security.celery:'
                        f'Celery task invocation "{logged_data.celery_task_invocation[0].id}" '
                        f'with celery id "{logged_data.celery_task_invocation[0].celery_task_id}" '
                        f'and name "sum_task" was successful',
                    ]
                )

    @override_settings(SECURITY_BACKEND_WRITERS={'sql'})
    def test_error_celery_task_should_be_logged_in_sql_backend(self):
        error_task.apply()
        assert_equal(SQLCeleryTaskInvocationLog.objects.count(), 1)
        assert_equal(SQLCeleryTaskRunLog.objects.count(), 1)
        sql_celery_task_invocation_log = SQLCeleryTaskInvocationLog.objects.get()
        sql_celery_task_run_log = SQLCeleryTaskRunLog.objects.get()

        assert_equal_model_fields(
            sql_celery_task_invocation_log,
            state=CeleryTaskInvocationLogState.FAILED,
            name='error_task',
            time=(sql_celery_task_invocation_log.stop - sql_celery_task_invocation_log.start).total_seconds(),
            expires_at=None,
            input='',
            task_args=[],
            task_kwargs={},
            queue_name='default'
        )
        assert_is_not_none(sql_celery_task_invocation_log.celery_task_id)
        assert_equal_model_fields(
            sql_celery_task_run_log,
            state=CeleryTaskRunLogState.FAILED,
            name='error_task',
            time=(sql_celery_task_run_log.stop - sql_celery_task_run_log.start).total_seconds(),
            input='',
            task_args=[],
            task_kwargs={},
            queue_name=None,
            retries=0
        )
        assert_is_not_none(sql_celery_task_run_log.error_message)

    @store_elasticsearch_log()
    def test_error_celery_task_should_be_logged_in_elasticsearch_backend(self):
        with capture_security_logs() as logged_data:
            error_task.apply()
            elasticsearch_celery_task_invocation_log = ElasticsearchCeleryTaskInvocationLog.get(
                id=logged_data.celery_task_invocation[0].id
            )
            ElasticsearchCeleryTaskRunLog._index.refresh()
            elasticsearch_celery_task_run_log = ElasticsearchCeleryTaskRunLog.get(
                id=logged_data.celery_task_run[0].id
            )

            assert_equal_model_fields(
                elasticsearch_celery_task_invocation_log,
                state=CeleryTaskInvocationLogState.FAILED,
                name='error_task',
                time=(
                    elasticsearch_celery_task_invocation_log.stop - elasticsearch_celery_task_invocation_log.start
                ).total_seconds(),
                expires_at=None,
                input='',
                task_args='[]',
                task_kwargs='{}',
                queue_name='default'
            )
            assert_is_not_none(elasticsearch_celery_task_invocation_log.celery_task_id)
            assert_equal_model_fields(
                elasticsearch_celery_task_run_log,
                state=CeleryTaskRunLogState.FAILED,
                name='error_task',
                time=(elasticsearch_celery_task_run_log.stop - elasticsearch_celery_task_run_log.start).total_seconds(),
                input='',
                task_args='[]',
                task_kwargs='{}',
                queue_name=None,
                retries=0
            )
            assert_is_not_none(elasticsearch_celery_task_run_log.error_message)

    @store_elasticsearch_log(SECURITY_ELASTICSEARCH_LOGSTASH_WRITER=True)
    def test_error_celery_task_should_be_logged_in_elasticsearch_backend_through_logstash(self):
        with capture_security_logs() as logged_data:
            with self.assertLogs('security.logstash', level='INFO') as cm:
                error_task.apply()
                celery_task_invocation_log = logged_data.celery_task_invocation[0]
                celery_task_run_log = logged_data.celery_task_run[0]

                assert_equal(len(cm.output), 5)
                start_invocation, trigger_invocation, start_run, failed_run, failed_invocation = cm.output

                start_invocation_expected_data = {
                    'slug': None,
                    'release': None,
                    'related_objects': [],
                    'extra_data': {},
                    'parent_log': None,
                    'name': 'error_task',
                    'queue_name': 'default',
                    'input': '',
                    'task_args': '[]',
                    'task_kwargs': '{}',
                    'applied_at': not_none_eq_obj,
                    'is_async': False,
                    'is_unique': False,
                    'is_on_commit': False,
                    'start': not_none_eq_obj,
                    'state': 'WAITING'
                }
                trigger_invocation_expected_data = {
                    **start_invocation_expected_data,
                    'triggered_at': not_none_eq_obj,
                    'state': 'TRIGGERED',
                    'stale_at': not_none_eq_obj,
                    'celery_task_id': not_none_eq_obj,
                    'estimated_time_of_first_arrival': not_none_eq_obj,
                    'expires_at': None,
                }
                failed_invocation_expected_data = {
                    **trigger_invocation_expected_data,
                    'state': 'FAILED'
                }

                start_run_expected_data = {
                    'slug': None,
                    'release': None,
                    'related_objects': [],
                    'extra_data': {},
                    'parent_log': None,
                    'name': 'error_task',
                    'celery_task_id': not_none_eq_obj,
                    'queue_name': None,
                    'input': '',
                    'task_args': '[]',
                    'task_kwargs': '{}',
                    'start': not_none_eq_obj,
                    'retries': 0,
                    'waiting_time': not_none_eq_obj,
                    'state': 'ACTIVE',
                }
                failed_run_expected_data = {
                    **start_run_expected_data,
                    'time': not_none_eq_obj,
                    'stop': not_none_eq_obj,
                    'state': 'FAILED',
                    'error_message': 'error'
                }

                assert_equal_logstash(
                    start_invocation,
                    'security-celery-task-invocation-log',
                    0,
                    celery_task_invocation_log.id,
                    start_invocation_expected_data
                )
                assert_equal_logstash(
                    trigger_invocation,
                    'security-celery-task-invocation-log',
                    1,
                    celery_task_invocation_log.id,
                    trigger_invocation_expected_data
                )

                assert_equal_logstash(
                    start_run,
                    'security-celery-task-run-log',
                    0,
                    celery_task_run_log.id,
                    start_run_expected_data
                )
                assert_equal_logstash(
                    failed_run,
                    'security-celery-task-run-log',
                    9999,
                    celery_task_run_log.id,
                    failed_run_expected_data
                )
                assert_equal_logstash(
                    failed_invocation,
                    'security-celery-task-invocation-log',
                    9999,
                    celery_task_invocation_log.id,
                    failed_invocation_expected_data
                )

    @override_settings(SECURITY_BACKEND_WRITERS={'logging'})
    def test_error_celery_task_should_be_logged_in_logging_backend(self):
        with capture_security_logs() as logged_data:
            with self.assertLogs('security.celery', level='INFO') as cm:
                error_task.apply()

                assert_equal(
                    cm.output,
                    [
                        f'INFO:security.celery:'
                        f'Celery task invocation "{logged_data.celery_task_invocation[0].id}" '
                        f'with celery id "{logged_data.celery_task_invocation[0].celery_task_id}" '
                        f'and name "error_task" was invoked',
                        f'INFO:security.celery:Celery task "{logged_data.celery_task_run[0].id}" with '
                        f'celery id "{logged_data.celery_task_run[0].celery_task_id}" and name "error_task" '
                        f'was started',
                        f'ERROR:security.celery:Celery task "{logged_data.celery_task_run[0].id}" with '
                        f'celery id "{logged_data.celery_task_run[0].celery_task_id}" and name "error_task" '
                        f'failed',
                        f'INFO:security.celery:'
                        f'Celery task invocation "{logged_data.celery_task_invocation[0].id}" '
                        f'with celery id "{logged_data.celery_task_invocation[0].celery_task_id}" '
                        f'and name "error_task" was failed',
                    ]
                )

    @override_settings(SECURITY_BACKEND_WRITERS={'sql'})
    def test_retry_celery_task_should_be_logged_in_sql_backend(self):
        with capture_security_logs() as logged_data:
            retry_task.apply()
            assert_equal(SQLCeleryTaskInvocationLog.objects.count(), 1)
            assert_equal(SQLCeleryTaskRunLog.objects.count(), 6)

            sql_celery_task_invocation_log = SQLCeleryTaskInvocationLog.objects.get()
            assert_equal_model_fields(
                sql_celery_task_invocation_log,
                state=CeleryTaskInvocationLogState.SUCCEEDED,
                name='retry_task',
                time=(sql_celery_task_invocation_log.stop - sql_celery_task_invocation_log.start).total_seconds(),
                expires_at=None,
                input='',
                task_args=[],
                task_kwargs={},
                queue_name='default'
            )

            for i, logger in enumerate(logged_data.celery_task_run[0:5]):
                sql_celery_task_run_log = SQLCeleryTaskRunLog.objects.get(id=logger.id)
                assert_equal_model_fields(
                    sql_celery_task_run_log,
                    state=CeleryTaskRunLogState.RETRIED,
                    name='retry_task',
                    time=(sql_celery_task_run_log.stop - sql_celery_task_run_log.start).total_seconds(),
                    input='',
                    task_args=[],
                    task_kwargs={},
                    queue_name=None,
                    retries=i
                )
                assert_is_not_none(sql_celery_task_run_log.error_message)
            sql_celery_task_run_log = SQLCeleryTaskRunLog.objects.get(id=logged_data.celery_task_run[-1].id)
            assert_equal_model_fields(
                sql_celery_task_run_log,
                state=CeleryTaskRunLogState.SUCCEEDED,
                name='retry_task',
                time=(sql_celery_task_run_log.stop - sql_celery_task_run_log.start).total_seconds(),
                input='',
                task_args=[],
                task_kwargs={},
                queue_name=None,
                retries=5,
                error_message=None
            )
            assert_equal(sql_celery_task_invocation_log.last_run.id, sql_celery_task_run_log.id)

    @store_elasticsearch_log()
    def test_retry_celery_task_should_be_logged_in_elasticsearch_backend(self):
        with capture_security_logs() as logged_data:
            retry_task.apply()

            elasticsearch_celery_task_invocation_log = ElasticsearchCeleryTaskInvocationLog.get(
                id=logged_data.celery_task_invocation[0].id
            )
            assert_equal_model_fields(
                elasticsearch_celery_task_invocation_log,
                state=CeleryTaskInvocationLogState.SUCCEEDED,
                name='retry_task',
                time=(
                    elasticsearch_celery_task_invocation_log.stop - elasticsearch_celery_task_invocation_log.start
                ).total_seconds(),
                expires_at=None,
                input='',
                task_args='[]',
                task_kwargs='{}',
                queue_name='default'
            )

            for i, logger in enumerate(logged_data.celery_task_run[0:5]):
                elasticsearch_celery_task_run_log = ElasticsearchCeleryTaskRunLog.get(
                    id=logger.id
                )
                assert_equal_model_fields(
                    elasticsearch_celery_task_run_log,
                    state=CeleryTaskRunLogState.RETRIED,
                    name='retry_task',
                    time=(
                        elasticsearch_celery_task_run_log.stop - elasticsearch_celery_task_run_log.start
                    ).total_seconds(),
                    input='',
                    task_args='[]',
                    task_kwargs='{}',
                    queue_name=None,
                    retries=i
                )
                assert_is_not_none(elasticsearch_celery_task_run_log.error_message)
            elasticsearch_celery_task_run_log = ElasticsearchCeleryTaskRunLog.get(
                id=logged_data.celery_task_run[-1].id
            )
            assert_equal_model_fields(
                elasticsearch_celery_task_run_log,
                state=CeleryTaskRunLogState.SUCCEEDED,
                name='retry_task',
                time=(elasticsearch_celery_task_run_log.stop - elasticsearch_celery_task_run_log.start).total_seconds(),
                input='',
                task_args='[]',
                task_kwargs='{}',
                queue_name=None,
                retries=5,
                error_message=None
            )
            ElasticsearchCeleryTaskRunLog._index.refresh()
            assert_equal(elasticsearch_celery_task_invocation_log.last_run.id, elasticsearch_celery_task_run_log.id)

    @store_elasticsearch_log(SECURITY_ELASTICSEARCH_LOGSTASH_WRITER=True)
    def test_retry_celery_task_should_be_logged_in_elasticsearch_backend_through_logstash(self):
        with capture_security_logs() as logged_data:
            with self.assertLogs('security.logstash', level='INFO') as cm:
                retry_task.apply()
                celery_task_invocation_log = logged_data.celery_task_invocation[0]

                assert_equal(len(cm.output), 15)
                start_invocation, trigger_invocation = cm.output[:2]

                start_invocation_expected_data = {
                    'slug': None,
                    'release': None,
                    'related_objects': [],
                    'extra_data': {},
                    'parent_log': None,
                    'name': 'retry_task',
                    'queue_name': 'default',
                    'input': '',
                    'task_args': '[]',
                    'task_kwargs': '{}',
                    'applied_at': not_none_eq_obj,
                    'is_async': False,
                    'is_unique': False,
                    'is_on_commit': False,
                    'start': not_none_eq_obj,
                    'state': 'WAITING'
                }
                trigger_invocation_expected_data = {
                    **start_invocation_expected_data,
                    'triggered_at': not_none_eq_obj,
                    'state': 'TRIGGERED',
                    'stale_at': None,
                    'celery_task_id': not_none_eq_obj,
                    'estimated_time_of_first_arrival': not_none_eq_obj,
                    'expires_at': None,
                }

                assert_equal_logstash(
                    start_invocation,
                    'security-celery-task-invocation-log',
                    0,
                    celery_task_invocation_log.id,
                    start_invocation_expected_data
                )
                assert_equal_logstash(
                    trigger_invocation,
                    'security-celery-task-invocation-log',
                    1,
                    celery_task_invocation_log.id,
                    trigger_invocation_expected_data
                )
                for i, celery_task_run_log in enumerate(logged_data.celery_task_run[0:5]):
                    start_run, retry_run = cm.output[2 + (i*2):4 + (i*2)]
                    start_run_expected_data = {
                        'slug': None,
                        'release': None,
                        'related_objects': [],
                        'extra_data': {},
                        'parent_log': None,
                        'name': 'retry_task',
                        'celery_task_id': not_none_eq_obj,
                        'queue_name': None,
                        'input': '',
                        'task_args': '[]',
                        'task_kwargs': '{}',
                        'start': not_none_eq_obj,
                        'retries': i,
                        'waiting_time': not_none_eq_obj,
                        'state': 'ACTIVE',
                    }
                    retry_run_expected_data = {
                        **start_run_expected_data,
                        'time': not_none_eq_obj,
                        'stop': not_none_eq_obj,
                        'state': 'RETRIED',
                        'error_message': 'error',
                        'estimated_time_of_next_retry': not_none_eq_obj,
                    }

                    assert_equal_logstash(
                        start_run,
                        'security-celery-task-run-log',
                        0,
                        celery_task_run_log.id,
                        start_run_expected_data
                    )
                    assert_equal_logstash(
                        retry_run,
                        'security-celery-task-run-log',
                        9999,
                        celery_task_run_log.id,
                        retry_run_expected_data
                    )
                celery_task_run_log = logged_data.celery_task_run[-1]
                start_run, succes_run = cm.output[12:14]
                start_run_expected_data = {
                    'slug': None,
                    'release': None,
                    'related_objects': [],
                    'extra_data': {},
                    'parent_log': None,
                    'name': 'retry_task',
                    'celery_task_id': not_none_eq_obj,
                    'queue_name': None,
                    'input': '',
                    'task_args': '[]',
                    'task_kwargs': '{}',
                    'start': not_none_eq_obj,
                    'retries': 5,
                    'waiting_time': not_none_eq_obj,
                    'state': 'ACTIVE',
                }
                succes_run_expected_data = {
                    **start_run_expected_data,
                    'time': not_none_eq_obj,
                    'stop': not_none_eq_obj,
                    'state': 'SUCCEEDED',
                    'result': None
                }

                assert_equal_logstash(
                    start_run,
                    'security-celery-task-run-log',
                    0,
                    celery_task_run_log.id,
                    start_run_expected_data
                )
                assert_equal_logstash(
                    succes_run,
                    'security-celery-task-run-log',
                    9999,
                    celery_task_run_log.id,
                    succes_run_expected_data
                )

                success_invocation = cm.output[14]
                success_invocation_expected_data = {
                    **trigger_invocation_expected_data,
                    'state': 'SUCCEEDED',
                }
                assert_equal_logstash(
                    success_invocation,
                    'security-celery-task-invocation-log',
                    9999,
                    celery_task_invocation_log.id,
                    success_invocation_expected_data
                )

    @override_settings(SECURITY_BACKEND_WRITERS={'logging'})
    def test_retry_celery_task_should_be_logged_in_logging_backend(self):
        with capture_security_logs() as logged_data:
            with self.assertLogs('security.celery', level='INFO') as cm:
                retry_task.apply()
                expected_output = [
                    f'INFO:security.celery:'
                    f'Celery task invocation "{logged_data.celery_task_invocation[0].id}" '
                    f'with celery id "{logged_data.celery_task_invocation[0].celery_task_id}" '
                    f'and name "retry_task" was invoked',
                ]
                for logger in logged_data.celery_task_run[0:5]:
                    expected_output += [
                        f'INFO:security.celery:Celery task "{logger.id}" with '
                        f'celery id "{logger.celery_task_id}" and name "retry_task" '
                        f'was started',
                        f'WARNING:security.celery:Celery task "{logger.id}" with '
                        f'celery id "{logger.celery_task_id}" and name "retry_task" '
                        f'was repeated',
                    ]
                expected_output += [
                    f'INFO:security.celery:Celery task "{logged_data.celery_task_run[-1].id}" with '
                    f'celery id "{logged_data.celery_task_run[-1].celery_task_id}" and name "retry_task" '
                    f'was started',
                    f'INFO:security.celery:Celery task "{logged_data.celery_task_run[-1].id}" with '
                    f'celery id "{logged_data.celery_task_run[-1].celery_task_id}" and name "retry_task" was '
                    f'successful',
                    f'INFO:security.celery:'
                    f'Celery task invocation "{logged_data.celery_task_invocation[0].id}" '
                    f'with celery id "{logged_data.celery_task_invocation[0].celery_task_id}" '
                    f'and name "retry_task" was successful',
                ]
                assert_equal(cm.output, expected_output)

    @override_settings(SECURITY_BACKEND_WRITERS={'sql'}, DJANGO_CELERY_EXTENSIONS_DEFAULT_TASK_STALE_TIME_LIMIT=30)
    def test_set_celery_task_log_state_should_set_task_to_failed_with_sql_backend(self):
        with mock.patch.object(unique_task, '_get_unique_task_id') as apply_method:
            unique_task_id = uuid4()
            apply_method.return_value = unique_task_id
            unique_task.apply_async()

            sql_celery_task_invocation_log = SQLCeleryTaskInvocationLog.objects.get()
            assert_equal(sql_celery_task_invocation_log.state, CeleryTaskInvocationLogState.DUPLICATE)

            test_call_command('set_celery_task_log_state')
            sql_celery_task_invocation_log.refresh_from_db()
            assert_equal(sql_celery_task_invocation_log.state, CeleryTaskInvocationLogState.DUPLICATE)

            change_and_save(sql_celery_task_invocation_log, state=CeleryTaskInvocationLogState.TRIGGERED, version=5)
            test_call_command('set_celery_task_log_state')
            assert_equal(sql_celery_task_invocation_log.state, CeleryTaskInvocationLogState.TRIGGERED)

            with freeze_time(now() + timedelta(seconds=30)):
                test_call_command('set_celery_task_log_state')
                sql_celery_task_invocation_log.refresh_from_db()
                assert_equal(sql_celery_task_invocation_log.state, CeleryTaskInvocationLogState.EXPIRED)

    @store_elasticsearch_log(DJANGO_CELERY_EXTENSIONS_DEFAULT_TASK_STALE_TIME_LIMIT=30)
    def test_set_celery_task_log_state_should_set_task_to_failed_with_elasticsearch_backend(self):
        with capture_security_logs() as logged_data:
            with mock.patch.object(unique_task, '_get_unique_task_id') as apply_method:
                unique_task_id = uuid4()
                apply_method.return_value = unique_task_id
                unique_task.apply_async()
                ElasticsearchCeleryTaskInvocationLog._index.refresh()

                test_call_command('set_celery_task_log_state')
                assert_equal(
                    ElasticsearchCeleryTaskInvocationLog.get(
                        id=logged_data.celery_task_invocation[0].id
                    ).state, CeleryTaskInvocationLogState.DUPLICATE
                )

                # Update index increase version over the maximum
                with mock.patch('security.backends.elasticsearch.writer.MAX_VERSION', 10001):
                    ElasticsearchCeleryTaskInvocationLog.get(
                        id=logged_data.celery_task_invocation[0].id
                    ).update(state=CeleryTaskInvocationLogState.TRIGGERED)
                    ElasticsearchCeleryTaskInvocationLog._index.refresh()
                    ElasticsearchCeleryTaskRunLog._index.refresh()

                    with freeze_time(now() + timedelta(seconds=30)):
                        test_call_command('set_celery_task_log_state')
                        assert_equal(
                            ElasticsearchCeleryTaskInvocationLog.get(
                                id=logged_data.celery_task_invocation[0].id
                            ).state,
                            CeleryTaskInvocationLogState.EXPIRED
                        )

    @override_settings(SECURITY_BACKEND_WRITERS={'sql'}, DJANGO_CELERY_EXTENSIONS_DEFAULT_TASK_STALE_TIME_LIMIT=30)
    def test_set_celery_task_log_state_should_set_task_to_succeeded_with_sql_backend(self):
        unique_task.apply_async()

        sql_celery_task_invocation_log = SQLCeleryTaskInvocationLog.objects.get()
        change_and_save(sql_celery_task_invocation_log, state=CeleryTaskInvocationLogState.TRIGGERED)

        test_call_command('set_celery_task_log_state')
        sql_celery_task_invocation_log.refresh_from_db()
        assert_equal(
            sql_celery_task_invocation_log.state, CeleryTaskInvocationLogState.TRIGGERED
        )

        with freeze_time(now() + timedelta(seconds=30)):
            test_call_command('set_celery_task_log_state')
            sql_celery_task_invocation_log.refresh_from_db()
            assert_equal(sql_celery_task_invocation_log.state, CeleryTaskInvocationLogState.SUCCEEDED)

    @store_elasticsearch_log(DJANGO_CELERY_EXTENSIONS_DEFAULT_TASK_STALE_TIME_LIMIT=30)
    def test_set_celery_task_log_state_should_set_task_to_succeeded_with_elasticsearch_backend(self):
        with capture_security_logs() as logged_data:
            unique_task.apply_async()
            ElasticsearchCeleryTaskInvocationLog.get(
                id=logged_data.celery_task_invocation[0].id
            ).update(state=CeleryTaskInvocationLogState.TRIGGERED)
            ElasticsearchCeleryTaskInvocationLog._index.refresh()
            ElasticsearchCeleryTaskRunLog._index.refresh()

            # Update index increase version over the maximum
            with mock.patch('security.backends.elasticsearch.writer.MAX_VERSION', 10001):
                test_call_command('set_celery_task_log_state')
                assert_equal(
                    ElasticsearchCeleryTaskInvocationLog.get(
                        id=logged_data.celery_task_invocation[0].id
                    ).state, CeleryTaskInvocationLogState.TRIGGERED
                )

                with freeze_time(now() + timedelta(seconds=30)):
                    test_call_command('set_celery_task_log_state')
                    assert_equal(
                        ElasticsearchCeleryTaskInvocationLog.get(
                            id=logged_data.celery_task_invocation[0].id
                        ).state,
                        CeleryTaskInvocationLogState.SUCCEEDED
                    )

    def test_acks_late_with_no_failure_or_timeout_acknowledgement_celery_task_should_not_log_invocation_error(self):
        expected_invocation_started_data = {
            'name': 'acks_late_with_no_failure_or_timeout_acknowledgement',
            'queue_name': 'default',
            'input': '',
            'task_args': [],
            'task_kwargs': {},
            'applied_at': not_none_eq_obj,
            'is_async': False,
            'is_unique': False,
            'is_on_commit': False,
            'start': not_none_eq_obj,
        }
        expected_invocation_triggered_data = {
            **expected_invocation_started_data,
            'triggered_at': not_none_eq_obj,
            'stale_at': None,
            'estimated_time_of_first_arrival': not_none_eq_obj,
            'expires_at': None,
            'celery_task_id': not_none_eq_obj,
        }

        expected_run_started_data = {
            'name': 'acks_late_with_no_failure_or_timeout_acknowledgement',
            'queue_name': None,
            'input': '',
            'task_args': [],
            'task_kwargs': {},
            'start': not_none_eq_obj,
            'retries': 0,
            'celery_task_id': not_none_eq_obj,
            'waiting_time': not_none_eq_obj,
        }
        expected_run_failed_data = {
            **expected_run_started_data,
            'stop': not_none_eq_obj,
            'error_message': not_none_eq_obj,
        }
        with capture_security_logs() as logged_data:
            acks_late_with_no_failure_or_timeout_acknowledgement.apply()
            assert_length_equal(logged_data.celery_task_invocation_started, 1)
            assert_length_equal(logged_data.celery_task_invocation_triggered, 1)
            assert_length_equal(logged_data.celery_task_invocation_failed, 0)
            assert_length_equal(logged_data.celery_task_run_started, 1)
            assert_length_equal(logged_data.celery_task_run_succeeded, 0)
            assert_length_equal(logged_data.celery_task_run_failed, 1)
            assert_length_equal(logged_data.celery_task_run_retried, 0)
            assert_equal_log_data(logged_data.celery_task_invocation_started[0], expected_invocation_started_data)

            assert_equal_log_data(logged_data.celery_task_invocation_triggered[0], expected_invocation_triggered_data)
            assert_equal_log_data(logged_data.celery_task_run_started[0], expected_run_started_data)
            assert_equal_log_data(logged_data.celery_task_run_failed[0], expected_run_failed_data)
